"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""

from PyQt5 import QtCore, QtWidgets

from .pages import BasesPage, CommoditiesPage, FactionsPage, ShipsPage, GunsPage, ThrustersPage, ArmourPage, IDsPage, \
                    CountermeasuresPage, ShieldsPage, MinesPage, CloaksPage, EnginesPage
from ...widgets.infocardview import InfocardView
from ...widgets.scrollablelist import ScrollableList


class Database(QtWidgets.QDialog):
    """The Database dialogue provides a spreadsheet-esque window into every entity defined in Freelancer. It more or
    less replicates the functionality of FLStat, using flint as a backend."""
    title = f'Database'
    defaultDimensions = (1400, 800)

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.title)
        self.resize(*self.defaultDimensions)
        self.setWindowFlags(QtCore.Qt.Window)

        self.pagesCache = {}

        self.mainLayout = QtWidgets.QHBoxLayout(self)

        viewSelector = ScrollableList(sorted(HEADINGS.keys()))
        viewSelector.setCurrentRow(1)
        viewSelector.currentTextChanged.connect(self.onSelectorChanged)
        self.mainLayout.addWidget(viewSelector, 0)

        self.infocardView = InfocardView(self)
        self.currentPage = BasesPage(self, self.infocardView)
        self.mainLayout.addWidget(self.currentPage, 4)
        self.mainLayout.addWidget(self.infocardView, 2)
        self.show()

    def onSelectorChanged(self, name):
        if HEADINGS[name] not in self.pagesCache:
            self.pagesCache[HEADINGS[name]] = HEADINGS[name](self, self.infocardView)
        newPage = self.pagesCache[HEADINGS[name]]

        self.currentPage.hide()
        self.mainLayout.replaceWidget(self.currentPage, newPage)
        self.currentPage = newPage
        self.currentPage.show()


HEADINGS = {
    'Bases': BasesPage,
    'Factions': FactionsPage,
    'Commodities': CommoditiesPage,
    'Ships': ShipsPage,
    'Indie IDs': IDsPage,
    'Guns': GunsPage,
    'Armour': ArmourPage,
    'Thrusters': ThrustersPage,
    'Mines': MinesPage,
    'CMs': CountermeasuresPage,
    'Shields': ShieldsPage,
    'Cloaks': CloaksPage,
    'Engines': EnginesPage,
}



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List, Iterator

from PyQt5 import QtCore, QtGui, QtWidgets

from ..models import items


class SimpleTree(QtWidgets.QTreeView):
    rowSelected = QtCore.pyqtSignal('PyQt_PyObject')  # emits a list of the cells in the selected row

    class FixedHeightDelegate(QtWidgets.QStyledItemDelegate):
        def sizeHint(self, option, index):
            size = super().sizeHint(option, index)
            size.setHeight(26)
            return size

    def __init__(self, header: List[str]):
        super().__init__()
        # configure model
        self.itemModel = QtGui.QStandardItemModel()
        self.itemModel.setHorizontalHeaderLabels(header)
        self.itemModel.setItemPrototype(items.GenericItem(''))
        self.setModel(self.itemModel)

        # configure view
        self.delegate = self.FixedHeightDelegate()
        self.setItemDelegate(self.delegate)

        self.setSortingEnabled(True)
        self.setHeaderHidden(False)
        self.setDropIndicatorShown(True)
        self.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.setDragEnabled(True)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.setUniformRowHeights(True)
        self.setAnimated(True)

    def allItems(self) -> Iterator[QtGui.QStandardItem]:
        root = self.itemModel.invisibleRootItem()

        def recurse(parent):
            for row in range(parent.rowCount()):
                for column in range(parent.columnCount()):
                    child = parent.child(row, column)
                    yield child
                    if child is not None and child.hasChildren():
                        yield from recurse(child)

        yield from recurse(root)

    def getSelectedRow(self) -> List[QtGui.QStandardItem]:
        return [i.model().sourceModel().itemFromIndex(i.model().mapToSource(i)) for i in self.selectedIndexes()]

    def selectionChanged(self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection):
        if not selected.indexes():
            return
        self.rowSelected.emit(self.getSelectedRow())



"""
Copyright © 2016-$today.year biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtWidgets
import flint as fl

from ...widgets.mapview import MapView


class ExpandedMap(MapView):
    """A windowed variant of MapView that can be used, for example, to display a universe map or an expanded
    system map."""
    INITIAL_SIZE = (800, 800)

    def __init__(self):
        super().__init__()
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.setWindowFlags(QtCore.Qt.Window)
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
        self.controlsFrame.hide()

    def heightForWidth(self, width: int) -> int:
        """Should always be square."""
        return width

    def display(self):
        """Display an expanded map."""
        self.resize(*self.INITIAL_SIZE)
        self.moveToCentre()
        self.show()

    def displayEntity(self, entity: fl.entities.Entity):
        """Display an expanded system map."""
        super().displayEntity(entity)
        self.setWindowModality(QtCore.Qt.NonModal)
        try:
            self.loadCompleted.disconnect()
        except TypeError:  # raised when no callbacks are connected
            pass
        self.loadCompleted.connect(lambda: self.setWindowTitle(self.getDisplayed()))
        self.display()

    def displayUniverse(self):
        """Display an expanded universe map."""
        super().displayUniverse()
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.resize(*self.INITIAL_SIZE)
        self.setWindowTitle('Sirius')
        self.show()
        try:
            self.loadCompleted.disconnect()
        except TypeError:  # raised when no callbacks are connected
            pass
        self.loadCompleted.connect(self.hide)

    def moveToCentre(self):
        """Move the widget to the centre of the currently active display."""
        geom = self.frameGeometry()
        desktop = QtWidgets.QApplication.desktop()
        activeDisplay = desktop.screenNumber(desktop.cursor().pos())
        centre = desktop.screenGeometry(activeDisplay).center()
        geom.moveCenter(centre)
        self.move(geom.topLeft())



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file contains the application's entry point - main().
"""
import sys

from PyQt5 import QtCore, QtGui, QtWidgets
import flint as fl

from . import app, config, icons, IS_WIN
from . import dialogues
from .windows.main import layout
from .dialogues import expandedmap, configurepaths
from .windows.main.navmap.navmap import Navmap
from .windows.main.merchant.merchant import Merchant
from .windows.main.roster.roster import Accounts

global mainWindow


# todo move paths dialog to own file

def onFirstRun():
    config['general']['firstrun'] = False


def main():
    """Main application entry point."""
    global mainWindow
    mainWindow = layout.MainWindow()

    firstRun = config['general'].getboolean('firstrun')

    # todo: get paths here

    if firstRun:
        onFirstRun()

    if IS_WIN:
        fl.paths.set_install_path(r"C:\Users\user\Desktop\Freelancer\Eclipse 2.1")
    else:
        fl.paths.set_install_path("/run/media/biqqles/Acer/Users/user/Desktop/Freelancer/Eclipse 2.1/")

    expmap = expandedmap.ExpandedMap()
    nav = Navmap(mainWindow, expmap)
    mer = Merchant(mainWindow, expmap, nav)
    acc = Accounts(mainWindow)

    mainWindow.dependents()

    sys.exit(app.exec())


if __name__ == '__main__':
    main()



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets

from ....widgets import buttons
from .... import icons

from ....widgets.simpletable import SimpleTable


class MerchantTab(QtWidgets.QWidget):
    """Defines the layout of the merchant tab."""
    title = 'Merchant'
    icon = icons.merchant

    def __init__(self, parent: QtWidgets.QTabWidget):
        super().__init__(parent)
        self.mainLayout = QtWidgets.QHBoxLayout(self)
        self.tableLayout = QtWidgets.QVBoxLayout()
        self.controlsLayout = QtWidgets.QHBoxLayout()
        self.originLabel = QtWidgets.QLabel('Origin', self)
        self.controlsLayout.addWidget(self.originLabel)

        self.originSelector = QtWidgets.QComboBox(self)
        self.originSelector.setToolTip('Origin system')
        self.originSelector.setEditable(True)
        self.originSelector.setMinimumHeight(30)
        self.controlsLayout.addWidget(self.originSelector)

        self.originMapButton = buttons.UniverseMapButton(self)
        self.controlsLayout.addWidget(self.originMapButton)

        self.controlsLayout.addSpacing(25)
        self.swapButton = buttons.SquareButton('⇆', parent=self)  # todo: turn this into an icon
        self.swapButton.setToolTip('Swap systems')
        font = QtGui.QFont()
        font.setPointSize(14)
        self.swapButton.setFont(font)
        self.controlsLayout.addWidget(self.swapButton)

        self.controlsLayout.addSpacing(25)

        self.destinationLabel = QtWidgets.QCheckBox('Destination', self)
        self.controlsLayout.addWidget(self.destinationLabel)

        self.destinationSelector = QtWidgets.QComboBox(self)
        self.destinationSelector.setToolTip('Destination system')  # todo make this combobox a widget
        self.destinationSelector.setEditable(True)
        self.destinationSelector.setEnabled(False)
        self.destinationSelector.lineEdit().setPlaceholderText('Anywhere')
        self.destinationSelector.setMinimumHeight(30)
        self.controlsLayout.addWidget(self.destinationSelector)

        self.destinationMapButton = buttons.UniverseMapButton(self)
        self.controlsLayout.addWidget(self.destinationMapButton)

        self.controlsLayout.addSpacing(40)

        self.advancedButton = QtWidgets.QPushButton('Advanced')
        self.advancedButton.setCheckable(True)
        self.controlsLayout.addWidget(self.advancedButton)

        self.controlsLayout.setStretch(1, 1)
        self.controlsLayout.setStretch(7, 1)
        self.tableLayout.addLayout(self.controlsLayout)

        # "additional controls" - since we want to selectively show/hide this layout it needs to be given a QWidget
        self.advancedControls = QtWidgets.QWidget()
        self.advancedControlsLayout = QtWidgets.QHBoxLayout()
        self.advancedControls.setLayout(self.advancedControlsLayout)

        self.contrabandCheck = QtWidgets.QCheckBox('Include contraband')
        self.advancedControlsLayout.addWidget(self.contrabandCheck)

        self.reputationLabel = QtWidgets.QCheckBox('Dockable by:')
        self.advancedControlsLayout.addWidget(self.reputationLabel)

        self.reputationSelector = QtWidgets.QComboBox()
        self.reputationSelector.setEnabled(False)
        self.reputationLabel.toggled.connect(self.reputationSelector.setEnabled)
        self.advancedControlsLayout.addWidget(self.reputationSelector)

        self.advancedControlsLayout.setAlignment(QtCore.Qt.AlignRight)

        self.tableLayout.addWidget(self.advancedControls)

        # button controls visibility; hidden by default
        self.advancedControls.hide()
        self.advancedButton.toggled.connect(self.advancedControls.setVisible)

        self.mainTable = SimpleTable(['Origin base', 'Origin system', 'Destination base', 'Destination system',
                                      'Commodity', 'Credits/unit'])
        self.mainTable.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.tableLayout.addWidget(self.mainTable)

        self.mainLayout.addLayout(self.tableLayout)

        self.infoLayout = QtWidgets.QVBoxLayout()
        self.infoLayout.setAlignment(QtCore.Qt.AlignHCenter)

        # a label which displays a commodity icon
        self.infoIcon = QtWidgets.QLabel()
        self.infoIcon.setFixedSize(128, 128)
        self.infoIcon.setScaledContents(True)
        self.infoIcon.setAlignment(QtCore.Qt.AlignCenter)
        self.infoLayout.addWidget(self.infoIcon, QtCore.Qt.AlignHCenter)

        self.infoNameLabel = QtWidgets.QLabel()
        self.infoNameLabel.setMaximumWidth(128)
        self.infoNameLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.infoNameLabel.setTextFormat(QtCore.Qt.RichText)
        self.infoNameLabel.setWordWrap(True)
        self.infoLayout.addWidget(self.infoNameLabel)

        self.infoBuyLabel = QtWidgets.QLabel()
        self.infoBuyLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.infoBuyLabel.setTextFormat(QtCore.Qt.RichText)
        self.infoBuyLabel.setWordWrap(True)
        self.infoLayout.addWidget(self.infoBuyLabel)

        self.infoSellLayout = QtWidgets.QLabel()
        self.infoSellLayout.setAlignment(QtCore.Qt.AlignCenter)
        self.infoSellLayout.setTextFormat(QtCore.Qt.RichText)
        self.infoSellLayout.setWordWrap(True)
        self.infoLayout.addWidget(self.infoSellLayout)
        self.infoLayout.addStretch(1)

        self.infoDivider = QtWidgets.QFrame()
        self.infoDivider.setFrameShape(QtWidgets.QFrame.HLine)
        self.infoLayout.addWidget(self.infoDivider)

        self.infoRouteLabel = QtWidgets.QLabel('Route')
        self.infoRouteLabel.setMaximumWidth(128)
        self.infoRouteLabel.setWordWrap(True)
        self.infoLayout.addWidget(self.infoRouteLabel)

        self.mainLayout.addLayout(self.infoLayout)



"""
Copyright © 2016-18 biqqles

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui


class TextFilter(QtCore.QSortFilterProxyModel):
    """A simple proxy model that filters on a configurable text query."""
    def __init__(self, sourceModel: QtGui.QStandardItemModel):
        super().__init__()
        self.setSourceModel(sourceModel)
        self.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.setDynamicSortFilter(True)
        self.setRecursiveFilteringEnabled(True)
        self.setFilterKeyColumn(-1)  # filter on all columns

        # delegate QStandardItemModel methods
        self.invisibleRootItem = sourceModel.invisibleRootItem
        self.itemFromIndex = sourceModel.itemFromIndex

    def update(self, query: str):
        """Cause the filter to be updated with a new query."""
        self.setFilterFixedString(query)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List, Tuple

from PyQt5 import QtCore, QtGui, QtWidgets


class ScrollableList(QtWidgets.QListWidget):
    def __init__(self, entries: List[str]):
        super().__init__()
        self.addItems(entries)
        self.setCurrentRow(0)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.setMaximumWidth(self.sizeHintForColumn(0) + 10)  # fit width to contents (+ margin)

    def wheelEvent(self, event: QtGui.QWheelEvent):
        """Allow selected item to be scrolled with the mouse wheel."""
        degrees = event.angleDelta().y() // 8  # angle delta is in eighths of a degree
        steps = degrees // 15
        for _ in range(abs(steps)):
            new_index = self.currentRow() + (1 if steps < 0 else -1)
            if -1 < new_index < self.count():  # do not wrap around
                self.setCurrentRow(new_index)
            else:
                break
        event.accept()



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtWidgets

from .. import icons, IS_WIN


class SquareButton(QtWidgets.QPushButton):
    """A button whose width is fixed to its height, so that it expands to the maximum possible size while remaining
    square. """

    def __init__(self, text='', tooltip='', icon=None, dropdown=False, edge=36, parent=None):
        """Initialise the widget with parameters passed to constructor."""
        super().__init__()
        self.setText(text)
        self.setToolTip(tooltip)
        if icon:
            self.setIcon(icon)
        self.setParent(parent)

        policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.setSizePolicy(policy)
        self.setMinimumSize(edge, edge)

        if IS_WIN:  # symbols are weirdly small in Segoe UI
            self.setStyleSheet('QPushButton {font-size: 18px;}')

        if dropdown:
            # hide dropdown arrow indicator
            self.setStyleSheet('QPushButton { padding-right: -0.5px } QPushButton::menu-indicator { image: none }')

    def resizeEvent(self, event):
        """Handle a resize event so that the widget stays square."""
        self.setFixedWidth(self.height())


class UniverseMapButton(SquareButton):
    """A square push button displaying the "universe map" symbol."""

    def __init__(self, parent=None):
        """Initialise a SquareButton with some default values."""
        super().__init__(icon=icons.universemap, tooltip='Open universe map', parent=parent)
        # self.setFixedHeight(QtWidgets.QComboBox().height())


"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtWidgets

from ....widgets.simpletree import SimpleTree
from .... import icons


class RosterTab(QtWidgets.QWidget):
    title = 'Roster'
    icon = icons.roster

    def __init__(self, parent: QtWidgets.QTabWidget):
        super().__init__(parent)
        self.mainLayout = QtWidgets.QHBoxLayout(self)

        self.treeLayout = QtWidgets.QVBoxLayout()
        self.filterLayout = QtWidgets.QHBoxLayout()
        self.filterLayout.setSpacing(50)
        self.warningLabel = QtWidgets.QLabel(
            'Wingman can only track characters for changes while they are logged in and while it is running.')
        self.filterLayout.addWidget(self.warningLabel)
        self.searchLineEdit = QtWidgets.QLineEdit()
        self.searchLineEdit.setPlaceholderText('Filter')
        self.filterLayout.addWidget(self.searchLineEdit)
        self.treeLayout.addLayout(self.filterLayout)

        self.tree = SimpleTree(['Name', 'Base', 'System', 'Credits', 'Last active', 'Description'])
        self.treeLayout.addWidget(self.tree)

        self.mainLayout.addLayout(self.treeLayout)

        self.controlsLayout = QtWidgets.QVBoxLayout()
        self.controlsLayout.setSpacing(15)
        self.controlsLayout.addSpacing(30)

        self.manualGroup = QtWidgets.QGroupBox('Edit')
        self.manualGroup.setAlignment(QtCore.Qt.AlignHCenter)
        self.manualLayout = QtWidgets.QVBoxLayout()
        self.addCharButton = QtWidgets.QPushButton('Add character')
        self.addCharButton.setToolTip('Manually add a new character (double click to edit)')
        self.manualLayout.addWidget(self.addCharButton)
        self.editCharButton = QtWidgets.QPushButton('Edit character')
        self.editCharButton.setToolTip('Allow the selected character to be edited')
        self.manualLayout.addWidget(self.editCharButton)
        self.moveCharButton = QtWidgets.QPushButton('Move character')
        self.moveCharButton.setToolTip('Allow the selected character to be moved between accounts')
        self.manualLayout.addWidget(self.moveCharButton)
        self.removeCharButton = QtWidgets.QPushButton('Remove character')
        self.removeCharButton.setToolTip('Remove the selected character')
        self.manualLayout.addWidget(self.removeCharButton)
        self.manualGroup.setLayout(self.manualLayout)
        self.controlsLayout.addWidget(self.manualGroup)

        self.fileGroup = QtWidgets.QGroupBox('File')
        self.fileGroup.setAlignment(QtCore.Qt.AlignHCenter)
        self.fileLayout = QtWidgets.QVBoxLayout()
        self.reloadButton = QtWidgets.QPushButton('Reload accounts')
        self.reloadButton.setToolTip('Reload account information from the launcher')
        self.fileLayout.addWidget(self.reloadButton)
        self.importExportButton = QtWidgets.QPushButton('Export')
        self.fileLayout.addWidget(self.importExportButton)
        self.fileGroup.setLayout(self.fileLayout)
        self.controlsLayout.addWidget(self.fileGroup)

        self.viewGroup = QtWidgets.QGroupBox('View')
        self.viewGroup.setAlignment(QtCore.Qt.AlignHCenter)
        self.viewLayout = QtWidgets.QVBoxLayout()
        self.collapseButton = QtWidgets.QPushButton('Collapse all')
        self.collapseButton.clicked.connect(lambda *args: self.tree.collapseAll())
        self.viewLayout.addWidget(self.collapseButton)
        self.expandButton = QtWidgets.QPushButton('Expand all')
        self.expandButton.clicked.connect(lambda *args: self.tree.expandAll())
        self.viewLayout.addWidget(self.expandButton)
        self.viewGroup.setLayout(self.viewLayout)
        self.controlsLayout.addWidget(self.viewGroup)

        self.controlsLayout.addStretch(1)

        self.mainLayout.addLayout(self.controlsLayout)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file defines the behaviour of the Navmap tab.
"""
import flint as fl
from PyQt5 import QtCore, QtWidgets

from ....widgets import mapview
from ...boxes import expandedmap
from .... import config, IS_WIN
from .layout import NavmapTab

if IS_WIN:
    import flair


class Navmap:
    """Implements the 'Navmap' tab."""
    def __init__(self, widget: NavmapTab, expandedMap: expandedmap.ExpandedMap):
        """Initialise tab"""
        self.config = config['navmap']
        self.searchableEntities = fl.systems + fl.bases
        self.currentlyDisplayed = self.searchableEntities.get(self.config['last'], self.searchableEntities['li01'])
        self.widget = widget
        self.mapView: mapview.MapView = self.widget.navmap
        self.expandedMap = expandedMap

        # set up search field with completer
        completer = QtWidgets.QCompleter()
        completer.setModel(QtCore.QStringListModel(e.name() for e in self.searchableEntities))
        completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        completer.setWrapAround(True)
        self.widget.searchEdit.setCompleter(completer)

        # connections
        self.mapView.loadCompleted.connect(self.onURLChange)
        self.mapView.loadCompleted.connect(print)
        self.widget.searchEdit.textEdited.connect(self.onSearchTextEdited)
        # textEdited is only emitted on user input, but completer counts as programmatic
        completer.activated.connect(self.onSearchTextEdited)
        completer.highlighted.connect(self.onSearchTextEdited)
        # navmap buttons
        self.mapView.backB.clicked.connect(self.widget.gotoRadioButton.click)
        self.mapView.forwardB.clicked.connect(self.widget.gotoRadioButton.click)
        self.mapView.expandB.clicked.connect(self.displayExpandedMap)
        self.widget.universeButton.clicked.connect(self.displayUniverseMap)

        self.mapView.setDisplayed(self.currentlyDisplayed.name())
        self.onURLChange(self.currentlyDisplayed.nickname)

        if IS_WIN:
            self.widget.followRadioButton.setEnabled(flair.state.running)
            flair.events.freelancer_started.connect(lambda: self.widget.followRadioButton.setEnabled(True))
            flair.events.freelancer_started.connect(lambda: self.widget.followRadioButton.setChecked(True))
            flair.events.freelancer_stopped.connect(lambda: self.widget.followRadioButton.setEnabled(False))
            flair.events.freelancer_stopped.connect(lambda: self.widget.gotoRadioButton.setChecked(True))
            flair.events.system_changed.connect(self.onFlairSystemChanged)
            self.widget.followRadioButton.toggled.connect(self.onFollowModeEnabled)

    def onURLChange(self, nickname):
        self.currentlyDisplayed = self.searchableEntities.get(nickname)
        self.displayInfocard(nickname)
        if nickname in self.searchableEntities:
            self.widget.searchEdit.setText(self.searchableEntities[nickname].name())  # update search field
            if nickname in fl.systems:
                system = fl.systems[nickname]
                self.searchableEntities += system.contents()  # load system contents
                self.mapView.displayConnMenu(system)
            self.config['last'] = nickname
        self.mapView.executeJS('wingman.onDisplayChanged()')

    def onSearchTextEdited(self, query: str):
        """Handle the search field's text being edited by the user."""
        self.widget.gotoRadioButton.setChecked(True)
        self.mapView.displayMap(query)

    def onFollowModeEnabled(self):
        """Handle follow mode being enabled."""
        if flair.state.system is not None:
            self.mapView.displayMap(flair.state.system)

    def onFlairSystemChanged(self, system: str):
        """Handle a system_changed event from flair."""
        if self.widget.followRadioButton.isChecked():
            self.mapView.displayMap(system)

    def displayInfocard(self, subject: str):
        """Displays an infocard for the given subject, where subject is a nickname"""
        if not subject:
            return

        self.widget.infocard.clear()

        if subject in self.searchableEntities:
            infocard = self.searchableEntities[subject].infocard().strip().rstrip('<p>')
            self.widget.infocard.append(infocard)
        else:
            self.widget.infocard.append('<i>No infocard available.</i>')

        self.widget.infocard.append(f'<hr><small>nickname: {subject}</small>')

        self.widget.infocard.verticalScrollBar().setValue(0)  # scroll to top

    def displayExpandedMap(self):
        """Display an expanded map of the current system."""
        self.expandedMap.displayEntity(self.currentlyDisplayed)

    def displayUniverseMap(self):
        """Display an expanded universe map. Selecting a system will display it in the main navmap."""
        self.expandedMap.displayUniverse()
        self.expandedMap.loadCompleted.connect(lambda n: self.mapView.displayEntity(fl.systems[n]))
        self.widget.gotoRadioButton.setChecked(True)

    def showFromExternal(self, entity: fl.entities.Entity):
        """Switch to the Navmap tab and show `entity` on the map."""
        self.mapView.displayEntity(entity)
        self.widget.parent().setCurrentWidget(self.widget)  # todo: doesn't update tab bar



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Any, TypeVar, Optional

from PyQt5 import QtGui, QtCore
from dataclassy import dataclass
import flint as fl
import ago


T = TypeVar('T')


class GenericItem(QtGui.QStandardItem):
    """A generic item with no data beyond its text."""
    def __init__(self, initialData: T):
        super().__init__()
        self.putData(initialData)
        self.setEditable(False)
        self.setDragEnabled(False)

    def putData(self, data: T):
        """An abstraction over setData(). Sets data and updates the textual representation."""
        self.setData(self.represent(data), QtCore.Qt.DisplayRole)
        self.setData(data, QtCore.Qt.UserRole)

    def getData(self) -> T:
        """Retrieve this item's underlying data."""
        return self.data(QtCore.Qt.UserRole)

    @staticmethod
    def represent(data: T) -> str:
        """Represent this item's data as a string. The default implementation returns the data unchanged, i.e. assumes
        it is a string to begin with.

        The "official" way to do this sort of thing is to do it all in a subclassed data(). However I find this less
        declarative than I'd like."""
        return data

    def serialise(self) -> Any:
        """Reduce this item's data down to a primitive that can be serialised in a format such as JSON."""
        return self.getData()

    @classmethod
    def deserialise(cls, serialised: Any):
        """Construct a new item from deserialised data. In most cases, this is simply equivalent to initialising
        the class normally."""
        return cls(serialised)

    def data(self, role=QtCore.Qt.UserRole + 1):
        """Alert the QItemDelegate to edit this item's underlying data."""
        if role == QtCore.Qt.EditRole:
            return self.getData()
        return super().data(role)

    def setData(self, value: float, role=QtCore.Qt.UserRole + 1):
        """Allow edits to update the underlying data."""
        if role == QtCore.Qt.EditRole:
            self.putData(value)
        else:
            super().setData(value, role)

    def clone(self) -> 'GenericItem':
        """Allow QStandardItemModel to clone this item, which happens e.g. on move events."""
        return GenericItem(self.getData())


class BlankItem(GenericItem):
    """A blank item."""
    def __init__(self):
        super().__init__('')


class MonospaceItem(GenericItem):
    def __init__(self, data):
        super().__init__(data)
        self.setFont(QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont))


class EntityItem(GenericItem):
    """An item holding a flint Entity."""
    @staticmethod
    def represent(entity: fl.entities.Entity):
        return entity.name()

    def data(self, role=QtCore.Qt.UserRole + 1):
        """Entities are never editable, but widgets such as QComboBox in text mode use the edit role."""
        if role == QtCore.Qt.EditRole:
            return self.text()
        return super().data(role)


class SystemItem(EntityItem):
    """An item holding a flint System."""
    def __init__(self, system: fl.entities.System):
        super().__init__(system)
        self.setToolTip(f'Region: {system.region()}')


class BaseItem(EntityItem):
    """An item holding a flint Base."""
    def __init__(self, base: fl.entities.Base):
        super().__init__(base)
        self.setToolTip(f'Sector: {base.sector()}\nIFF: {base.owner().name()}')


class FactionItem(EntityItem):
    """An item holding a flint Faction."""
    def __init__(self, faction: fl.entities.Faction):
        super().__init__(faction)


class CommodityItem(EntityItem):
    """An item holding a flint Commodity."""
    def __init__(self, commodity: fl.entities.Good):
        super().__init__(commodity)
        # self.setIcon(QtGui.QIcon(QtGui.QPixmap.fromImage(QtGui.QImage.fromData(commodity.icon(), 'TGA'))))


class NumberItem(GenericItem):
    """An item holding a number."""
    @staticmethod
    def represent(number):
        return f'{number:,}'

    def __lt__(self, other):
        return self.getData() < other.getData()


class BooleanItem(NumberItem):
    """An item displaying a boolean state."""
    def __init__(self, checked: bool):
        super().__init__(int(checked))
        self.setText('')
        self.setCheckState(QtCore.Qt.Checked if checked else QtCore.Qt.Unchecked)

    def setEditable(self, editable: bool):
        """Override set editable to allow the checkbox to be toggled but not the text to be edited."""
        self.setCheckable(editable)
        super().setEditable(False)


class PercentageItem(NumberItem):
    """An item displaying a number as a percentage."""
    @staticmethod
    def represent(number):
        return f'{number * 100:.2f}%'


class CreditsItem(NumberItem):
    """An item displaying a number as an amount in credits."""
    @staticmethod
    def represent(credits_):
        return f'${credits_:,}'


class ProfitItem(CreditsItem):
    """An item holding information about the profit of a transaction. It displays like a CreditsItem."""
    @dataclass
    class ProfitData:
        """Data for this transaction."""
        buyPrice: int
        buyBase: fl.entities.Base
        sellPrice: int
        sellBase: fl.entities.Base
        commodity: fl.entities.Commodity

        def profit(self):
            return (self.sellPrice - self.buyPrice) // self.commodity.volume

    def __init__(self, *args):
        data = self.ProfitData(*args)
        super().__init__(data)
        self.setToolTip(f'Buy: ${data.buyPrice}\nSell: ${data.sellPrice}\nVolume: {data.commodity.volume}')

    @staticmethod
    def represent(profitData: ProfitData):
        return f'${profitData.profit():,}'

    def __lt__(self, other):
        return self.getData().profit() < other.getData().profit()


class AccountItem(GenericItem):
    @dataclass
    class Account:
        hash: str
        launcher_name: str
        launcher_description: str

    def __init__(self, account: Account):
        super().__init__(account)
        font = self.font()
        font.setBold(True)
        self.setFont(font)
        self.setDropEnabled(True)

    @staticmethod
    def represent(account: Account) -> str:
        return account.launcher_name

    def serialise(self) -> Any:
        return self.getData().hash


class DateItem(NumberItem):
    timer = QtCore.QTimer()
    timer.start(60_000)

    def __init__(self, initialDate: Optional[QtCore.QDateTime]):
        super().__init__(initialDate)
        self.timer.timeout.connect(self.tick)

    def tick(self):
        """Keep the the displayed time delta accurate."""
        self.setText(self.represent(self.getData()))

    @staticmethod
    def represent(dateTime: Optional[QtCore.QDateTime]):
        if dateTime is None:
            return 'Unknown'
        if abs(dateTime.secsTo(QtCore.QDateTime.currentDateTime())) < 30:
            return 'Now'
        return ago.human(dateTime.toPyDateTime(), precision=1)

    def serialise(self) -> int:
        try:
            return self.getData().toSecsSinceEpoch()
        except AttributeError:
            return 0

    @classmethod
    def deserialise(cls, serialised: int):
        return cls(QtCore.QDateTime.fromSecsSinceEpoch(serialised))

    def data(self, role=QtCore.Qt.UserRole + 1):
        """Alert the QItemDelegate to edit this item as a date."""
        if role == QtCore.Qt.EditRole:
            return self.getData().date()
        return super().data(role)

    def setData(self, date: QtCore.QDate, role=QtCore.Qt.UserRole + 1):
        """Confirm the date is valid before accepting it."""
        if role == QtCore.Qt.EditRole:
            if date.isValid():
                self.putData(QtCore.QDateTime(date))
        else:
            super().setData(date, role)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file defines the interface of the application's main window.
"""
from . import layout



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file defines the interface of the application's main window.
"""
from PyQt5 import QtCore, QtWidgets
from ... import IS_WIN
if IS_WIN:
    import flair


class FlairBanner(QtWidgets.QLabel):
    """A banner label which displays live information provided by flair."""
    def __init__(self, window: QtWidgets.QMainWindow):
        super().__init__(parent=window)
        self.window = window
        self.setTextFormat(QtCore.Qt.RichText)
        self.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing | QtCore.Qt.AlignVCenter)

        if IS_WIN:
            self.timer = QtCore.QTimer(self)
            self.timer.setInterval(self.UPDATE_INTERVAL)
            self.timer.timeout.connect(self.updateContents)
            self.timer.start()

        self.updateContents()
        self.updatePosition()

    def updateContents(self):
        """Update the contents of the banner, using its template."""
        try:
            if not IS_WIN or not flair.state.running:
                self.hide()
            else:
                self.setText(self.TEMPLATE.format(flair.state.name, flair.state.system, flair.state.credits or 0))
                self.show()
        except AttributeError:  # flair has not been initialised yet
            self.hide()
            
    def updatePosition(self):
        """Update the position of the banner so that it "sticks" to the top right of the window."""
        tabBarBottom = self.window.tw.tabBar().tabRect(0).bottom()
        self.setFixedWidth(self.window.width() // 2)
        self.move(self.window.width() - self.X_OFFSET - self.width(), tabBarBottom - self.Y_OFFSET)

    X_OFFSET = 10  # offset from edge of screen
    Y_OFFSET = 4  # offset from top of tab widget
    UPDATE_INTERVAL = 1500  # ms
    TEMPLATE = ('&nbsp;' * 6).join(('Character: <i>{}</i>', 'System: <i>{}</i>', 'Credits: <i>${:,}</i>'))



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Tuple, Optional
from collections import defaultdict
from io import BytesIO

from PyQt5 import QtCore, QtGui, QtWidgets
from PIL import Image, ImageQt
import flint as fl

from .... import config
from ...boxes import expandedmap
from ....models import items, selectors
from .layout import MerchantTab


class Merchant:
    """Contains methods related to the 'Merchant' tab."""
    def __init__(self, widget: MerchantTab, expandedMap: expandedmap.ExpandedMap, navmap):
        # initialise Merchant tab
        self.config = config['merchant']
        self.widget = widget
        self.expandedMap = expandedMap
        self.navmap = navmap
        self.lastDestination: Optional[fl.entities.System] = None

        # create models
        systemsModel = selectors.SystemSelectionModel()
        self.widget.originSelector.setModel(systemsModel)
        self.widget.originMapButton.clicked.connect(lambda:
                                                    self.openUniverseMap(self.widget.originSelector))
        self.widget.destinationSelector.setModel(systemsModel)
        self.widget.destinationMapButton.clicked.connect(lambda:
                                                         self.openUniverseMap(self.widget.destinationSelector))

        self.widget.reputationSelector.setModel(selectors.FactionSelectionModel())

        # connections
        self.widget.originSelector.currentIndexChanged.connect(self.display)
        self.widget.destinationLabel.toggled.connect(self.onDestinationToggled)
        self.widget.destinationSelector.currentIndexChanged.connect(self.display)
        self.widget.reputationLabel.clicked.connect(self.display)
        self.widget.reputationSelector.currentIndexChanged.connect(self.display)
        self.widget.swapButton.clicked.connect(self.swapSystems)

        # customise table and model
        self.widget.mainTable.doubleClicked.connect(lambda index: self.displayBaseInNavmap(index))
        self.widget.mainTable.selectionModel().selectionChanged.connect(self.onSelectedRowChanged)

        # finally, display last settings
        self.selectedSystems = \
            (items.fl.systems.get(self.config['lastOrigin']), items.fl.systems.get(self.config['lastDestination']))

    def openUniverseMap(self, selector: QtWidgets.QComboBox):
        """Open the universe map, passing its choice to the given combo box."""
        self.expandedMap.displayUniverse()
        self.expandedMap.loadCompleted.connect(lambda n: selector.setCurrentIndex(selector.findData(items.fl.systems[n])))

    def onDestinationToggled(self, toggled: bool):
        """Handle the specific destination check box being toggled."""
        self.widget.destinationSelector.setEnabled(toggled)
        self.widget.destinationMapButton.setEnabled(toggled)
        if not toggled:
            self.lastDestination = self.widget.destinationSelector.currentData()
            self.widget.destinationSelector.setCurrentIndex(-1)
        elif self.lastDestination:
            self.widget.destinationSelector.setCurrentIndex(
                self.widget.destinationSelector.findData(self.lastDestination))
        self.display()

    def onSelectedRowChanged(self, selected, deselected):
        """Handle the table's selected row being changed."""
        if not selected.indexes():
            return
        originBase, originSystem, destinationBase, destinationSystem, commodity, profit = \
            (item.data(QtCore.Qt.UserRole) for item in selected.indexes())

        # display commodity icon
        self.updateInfoPanel(profit)

        # display commodity name
        links = []
        for system in items.fl.maps.inter_system_route(originBase._system, destinationBase._system):
            links.append(f'<a href="{system.nickname}">{system.name()}</a>')
        self.widget.infoRouteLabel.setText('Route: ' + ', '.join(links))
        self.widget.infoRouteLabel.linkActivated.connect(lambda n: self.navmap.showFromExternal(items.fl.systems[n]))

    def updateInfoPanel(self, data: items.ProfitItem.ProfitData):
        """Update the info side panel."""
        # load and display commodity icon
        icon = data.commodity.icon()
        image = QtGui.QImage.fromData(icon, 'TGA')
        # some icons fail to load into QImages with PyQt5 itself. I am yet to ascertain why. When this happens we
        # can use Pillow to load the file instead since they work fine with it. 0x3039 seems to be the first pixel value
        # on failure to load, don't know why.
        if image.pixel(QtCore.QPoint(0, 0)) == 0x3039:
            image = ImageQt.ImageQt(Image.open(BytesIO(icon)))
        pixmap = QtGui.QPixmap.fromImage(image)
        self.widget.infoIcon.setPixmap(pixmap)

        # update labels
        self.widget.infoNameLabel.setText(f'<b>{data.commodity.name()}</b>')
        self.widget.infoBuyLabel.setText(f'<b>Buy: ${data.buyPrice:,}</b><br/>{data.buyBase.name()}')
        self.widget.infoSellLayout.setText(f'<b>Sell: ${data.sellPrice:,}</b><br/>{data.sellBase.name()}')

    def display(self):
        """Display a table of the best commodities to take to and from the selected systems."""
        originSystem = self.widget.originSelector.itemData(self.widget.originSelector.currentIndex())
        destinationSystem = self.widget.destinationSelector.itemData(self.widget.destinationSelector.currentIndex()) if \
            self.widget.destinationLabel.isChecked() else None
        iff = self.widget.reputationSelector.itemData(self.widget.reputationSelector.currentIndex()) if \
            self.widget.reputationLabel.isChecked() else None

        data = self.calculateRoutes(originSystem, destinationSystem, iff)
        self.widget.mainTable.populate(list(data))
        self.widget.mainTable.sortByColumn(5, QtCore.Qt.DescendingOrder)  # sort by 'credits/unit' column
        self.widget.mainTable.selectRow(0)

        self.config['lastOrigin'] = originSystem.nickname
        self.config['lastDestination'] = destinationSystem.nickname if destinationSystem else ''

    def swapSystems(self):
        """Swap the currently selected origin and destination systems."""
        self.selectedSystems = reversed(self.selectedSystems)

    @property
    def selectedSystems(self):
        """Return the entities of the selected origin and destination system, respectively."""
        return self.widget.originSelector.currentText(), self.widget.destinationSelector.currentText()

    @selectedSystems.setter
    def selectedSystems(self, newSystems: Tuple[items.fl.entities.System, items.fl.entities.System]):
        origin, destination = newSystems
        if origin:
            self.widget.originSelector.setCurrentIndex(self.widget.originSelector.findData(origin))
        if destination:
            self.widget.destinationSelector.setCurrentIndex(self.widget.destinationSelector.findData(destination))
        else:
            self.widget.destinationSelector.setCurrentText('')

    def onUserDisplayRequested(self, origin, destination):
        self.setSystems(origin, destination)
        self.widget.tw.setCurrentIndex(1)

    @staticmethod
    def calculateRoutes(origin: items.fl.entities.System, destination: items.Optional[items.fl.entities.System] = None,
                        iff: items.Optional[items.fl.entities.Faction] = None):

        originSells = defaultdict(list)
        for base in origin.bases():
            if iff and not iff.can_dock_at(base):
                continue
            for good, price in base.universe_base().sells().items():
                originSells[good].append((price, base))

        if destination:
            destinationBases = destination.bases()
        else:
            destinationBases = {b for s in items.fl.systems for b in s.bases()}

        destBuys = defaultdict(list)
        for base in destinationBases:
            if iff and not iff.can_dock_at(base):
                continue
            for good, price in base.universe_base().buys().items():
                destBuys[good].append((price, base))

        for good in originSells:  # for each commodity that are sold in the origin system
            if good not in destBuys:  # if somewhere buys the commodity in the destination
                continue
            for destPrice, destBase in destBuys[good]:  # look through each base in destination
                for originPrice, originBase in originSells[good]:  # look through each base in origin
                    if destPrice > originPrice:
                        yield [items.BaseItem(originBase), items.SystemItem(originBase.system()),
                               items.BaseItem(destBase), items.SystemItem(destBase.system()),
                               items.CommodityItem(good.commodity()),
                               items.ProfitItem(originPrice, originBase, destPrice, destBase, good.commodity())]



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
import os.path
from pathlib import Path
import sys

from PyQt5 import QtCore, QtGui, QtWidgets
import flint as fl

from ...widgets.pathedit import PathEdit
from ... import IS_WIN, config


class ConfigurePaths(QtWidgets.QDialog):
    """A dialogue that allows the paths needed by the application to be configured."""
    title = 'Configure paths'

    def __init__(self, mandatory=False):
        super().__init__()
        self.mandatory = mandatory
        self.setWindowTitle(self.title)
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.setWindowFlags(QtCore.Qt.Dialog)
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
        self.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Maximum)

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.setLayout(self.mainLayout)

        self.formLayout = QtWidgets.QFormLayout()
        self.mainLayout.addLayout(self.formLayout)

        defaultFreelancerDir = config.paths['freelancer_dir'] or \
            os.path.expanduser('~')
        self.freelancerDirEdit = PathEdit(defaultFreelancerDir, fl.paths.is_probably_freelancer)
        self.freelancerDirEdit.lineEdit.textChanged.connect(self.validate)
        self.formLayout.addRow('Freelancer directory', self.freelancerDirEdit)

        defaultMyGamesDir = config.paths['my_games'] or \
            (os.path.expanduser(r'~\Documents\My Games' if IS_WIN else '~'))
        self.myGamesDirEdit = PathEdit(defaultMyGamesDir, lambda path: Path(path).name == 'My Games')
        self.myGamesDirEdit.lineEdit.textChanged.connect(self.validate)
        self.formLayout.addRow("'My Games' directory", self.myGamesDirEdit)

        self.buttons = QtWidgets.QDialogButtonBox()
        self.buttons.addButton(QtWidgets.QDialogButtonBox.Save)
        self.buttons.addButton(QtWidgets.QDialogButtonBox.Cancel)
        self.buttons.accepted.connect(self.writeConfig)
        self.buttons.rejected.connect(self.close)
        self.mainLayout.addWidget(self.buttons)

        self.show()
        self.validate()

    def validate(self):
        """Validate the provided paths, enabling or disabling the accept role."""
        self.buttons.button(QtWidgets.QDialogButtonBox.Save).setEnabled(self.allPathsValid())

    def allPathsValid(self):
        """Returns whether all paths entered are valid."""
        return self.freelancerDirEdit.isValidPath() and self.myGamesDirEdit.isValidPath()

    def writeConfig(self):
        """Commit the changes to the config file to disk. It should only be possible to call this after the paths
        have been confirmed as valid."""
        config.paths['freelancer_dir'] = self.freelancerDirEdit.lineEdit.text()
        config.paths['my_games'] = self.myGamesDirEdit.lineEdit.text()
        config.commit()
        self.close()

    def closeEvent(self, event: QtGui.QCloseEvent):
        """If the dialogue is mandatory, quit the application."""
        if self.mandatory and not self.allPathsValid():
            sys.exit(1)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file defines the interface of the application's main window.
"""
from typing import Callable, Union
import logging

from PyQt5 import QtCore, QtWidgets
from flint import cached

from ... import config, __app__
from . import merchant, navmap, roster, banner, menus
from .navmap.navmap import Navmap
from .merchant.merchant import Merchant
from .roster.roster import Roster
from ...windows.boxes.expandedmap import ExpandedMap


class MainWindow(QtWidgets.QMainWindow):
    """The application's main window."""
    title = __app__
    initialDimensions = (1060, 610)

    def __init__(self):
        logging.info('Loading main window')
        super().__init__()
        self.setWindowTitle(self.title)
        self.resize(*self.initialDimensions)

        self.centralWidget = QtWidgets.QWidget()
        self.centralLayout = QtWidgets.QVBoxLayout(self.centralWidget)
        self.centralLayout.setContentsMargins(0, 0, 0, 0)

        self.setCentralWidget(self.centralWidget)

        self.tw = QtWidgets.QTabWidget(self.centralWidget)
        self.tw.setContentsMargins(0, 0, 0, 0)

        # Navmap tab

        self.tabMap = navmap.layout.NavmapTab(self.tw)
        self.tw.addTab(self.tabMap, self.tabMap.icon, self.tabMap.title)
        self.tw.currentChanged.connect(self.cueLazyLoadTab(self.tabMap, self.navmap))

        # Merchant tab

        self.tabMer = merchant.layout.MerchantTab(self.tw)
        self.tw.addTab(self.tabMer, self.tabMer.icon, self.tabMer.title)
        self.tw.currentChanged.connect(self.cueLazyLoadTab(self.tabMer, self.merchant))

        # Roster

        self.tabRoster = roster.layout.RosterTab(self.tw)
        self.tw.addTab(self.tabRoster, self.tabRoster.icon, self.tabRoster.title)
        self.tw.currentChanged.connect(self.cueLazyLoadTab(self.tabRoster, self.roster))

        self.centralLayout.addWidget(self.tw)

        # banner

        self.banner = banner.FlairBanner(self)

        # Menubar

        self.menubar = QtWidgets.QMenuBar()
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1060, 35))
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.menubar.sizePolicy().hasHeightForWidth())
        self.menubar.setSizePolicy(sizePolicy)
        # self.menubar.setNativeMenuBar(True)

        self.setMenuBar(self.menubar)

        self.menubar.addMenu(menus.Utilities(self.menubar))
        self.menubar.addMenu(menus.File(self.menubar))
        self.menubar.addMenu(menus.Freelancer(self.menubar))
        self.menubar.addMenu(menus.Preferences(self.menubar))
        self.menubar.addMenu(menus.Help(self.menubar))

        self.show()
        self.tw.currentChanged.emit(0)

    def resizeEvent(self, event):
        """Ensure banner "sticks" to the top right of the window."""
        self.banner.updatePosition()

    def closeEvent(self, event):
        """Write config to disk on close."""
        config.commit()

    def cueLazyLoadTab(self, layout, load: Callable[[], Union[Navmap, Merchant, Roster]]):
        """Cue a tab to be lazy loaded."""
        return lambda: load() if self.tw.currentWidget() is layout else None

    @cached  # bit of a hack - using flint's decorator
    def expandedMap(self) -> ExpandedMap:
        """Initialise the expanded map."""
        return ExpandedMap()

    @cached
    def navmap(self) -> Navmap:
        """Load Navmap."""
        return Navmap(self.tabMap, self.expandedMap())

    @cached
    def merchant(self) -> Merchant:
        """Load Merchant."""
        return Merchant(self.tabMer, self.expandedMap(), self.navmap())

    @cached
    def roster(self) -> Roster:
        """Load Roster."""
        return Roster(self.tabRoster)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file initialises the application and contains key utility
classes.
"""
import configparser
import os
import sys

from PyQt5 import QtCore, QtGui, QtWidgets, QtWebEngineWidgets  # WebEngine must be imported here before QApp init

from . import resources


__version__ = '1.0.1 Private'


class Configuration(configparser.ConfigParser):
    """A class providing persistent configuration ... All data is stored on memory until program shutdown, when it is written to
    backing storage automatically."""
    path = os.path.join('cache', 'wingman.cfg')

    def __init__(self):
        configparser.ConfigParser.__init__(self, allow_no_value=True)
        self.optionxform = str  # override default behaviour of converting all keys to lowercase

        try:
            self.read_file(open(self.path))  # .read will not raise an error if the file doesn't exist
            assert {'paths', 'urls'}.issubset(self.sections())
        except FileNotFoundError or AssertionError:
            self.createDefault()
        self.createShortcuts()

    def save(self, *path):
        if not path:
            path = self.path

        with open(path, 'w') as fh:
            self.write(fh)

    def export(self, path):
        self.save(path)

    def createDefault(self):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)  # create cache directory if it doesn't already exist
        # load default file from resources
        file = QtCore.QFile(':/config/default')
        file.open(QtCore.QIODevice.ReadOnly | QtCore.QFile.Text)
        default = QtCore.QTextStream(file).readAll()
        self.read_string(default)
        self.save()

    def getlist(self, section, option):
        """Return a list stored as a newline delimited value.
        Analogous to the superclass methods getboolean, getint, etc."""
        value = self.get(section, option)
        return list(filter(None, (x.strip() for x in value.splitlines())))

    def createShortcuts(self):
        # todo move this to init and then reinit instead of calling this when nec
        # define some useful path constants
        paths = self['paths']
        self.INSTALL = paths['winstall'] if IS_WIN else paths['linstall']
        self.MYGAMES = paths['wmygames'] if IS_WIN else paths['lmygames']
        self.DSACE = self.fpath('Freelancer', 'DSAce.log', root=self.MYGAMES)
        self.LAUNCHERACCS = self.fpath('Discovery', 'launcheraccounts.xml', root=self.MYGAMES)
        self.NAVMAP = self['urls']['navmap']

    def fpath(self, *subpaths, root=None):
        """
        A convenient method which constructs an absolute from a given subpath, relative to root, which should be one of the path
        shortcuts mentioned above.
        """
        if not root:
            root = self.INSTALL
        return os.path.join(root, *subpaths)

    def createShared(self, path):
        """Create a symlink to the cache directory at `path`"""
        assert os.path.isdir('cache')
        os.rename('cache', '_cache')
        os.symlink(path, 'cache')

    def revertShared(self):
        assert os.path.islink('cache')
        os.remove('cache')
        os.rename('_cache', 'cache')


class Icons:
    """Set the icons of the interface depending on whether the background colour is light (meaning that dark icons should be used)
    or dark (meaning that light icons should be used)"""
    monoChrome = ['universemap', 'jump', 'expand', 'open', 'cc']

    def __init__(self):
        prefix = ':/dark/' if self.determineLuminescence() else ':/light/'

        for icon in self.monoChrome:
            setattr(self, icon, QtGui.QIcon(prefix + icon))

        self.main = QtGui.QIcon(':/general/main')  # todo: make static class
        self.navmap = QtGui.QIcon(':/general/navmap')
        self.merchant = QtGui.QIcon(':/general/merchant')
        self.roster = QtGui.QIcon(':/general/roster')

        # Linux only:
        self.open = QtGui.QIcon.fromTheme('document-save')
        self.copy = QtGui.QIcon.fromTheme('edit-copy')

    @staticmethod
    def determineLuminescence(threshold=200) -> bool:
        """Determine whether this platform has a light or dark theme by testing the background colour of a push
        button."""
        button = QtWidgets.QPushButton()
        colour = button.palette().color(button.backgroundRole())
        brightness = sum(colour.getRgb()) / 3
        return brightness > threshold


try:
    import flair
    IS_WIN = flair.IS_WIN
    IS_ADMIN = flair.IS_ADMIN
except ImportError:
    IS_WIN = False
    IS_ADMIN = False


plugin = QtCore.QPluginLoader('/home/biqqles/.local/lib/python3.7/site-packages/PyQt5/Qt/plugins/platformthemes/KDEPlasmaPlatformTheme.so')
print(plugin.load())
QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create('Style::Breeze'))


app = QtWidgets.QApplication(sys.argv)
app.setWindowIcon(QtGui.QIcon(':/general/main'))

# platform specific initialisation
if IS_WIN:
    font = app.font()
    font.setFamily('Segoe UI')
    font.setPointSizeF(font.pointSize() * 1.2)
    app.setFont(font)
from PyQt5 import Qt
print('version', Qt.QT_VERSION_STR)

    # switch cwd to directory and add to path (workaround for potential problems after running after install)
    # abspath = os.path.abspath(__file__)
    # dname = os.path.dirname(abspath)
    # os.chdir(dname)
    # sys.path.insert(0, dname)


icons = Icons()
config = Configuration()


if QtWidgets.QStyleFactory.keys() == ['Windows', 'Fusion']:
    print('Warning: native Qt style not available. Defaulting to Fusion - the application may not render correctly.'
          ' To fix this, install the PyQt5 packages from your distro\'s repos, not pip')



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtGui
import flint as fl

from .items import SystemItem, FactionItem


class SystemSelectionModel(QtGui.QStandardItemModel):
    def __init__(self):
        super().__init__()
        for system in sorted(fl.systems, key=lambda s: s.name()):
            if system.bases() and system.connections():
                self.appendRow(SystemItem(system))


class FactionSelectionModel(QtGui.QStandardItemModel):
    def __init__(self):
        super().__init__()
        for group in sorted(fl.factions, key=lambda f: f.name()):
            self.appendRow(FactionItem(group))



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file initialises the application.
"""
__app__ = 'Wingman'
__version__ = 'v0.1.2'
__description__ = 'A companion for Discovery Freelancer'

import os
import logging
import sys

# noinspection PyUnresolvedReferences
from PyQt5 import QtCore, QtGui, QtWidgets, QtWebEngineWidgets  # WebEngine must be imported here before QApp init
from . import resources  # register resources
from . import namespaces

try:
    # noinspection PyUnresolvedReferences
    import flair
    IS_WIN = True
except ImportError:
    IS_WIN = False

# paths relative to app data directory
NAVMAP_DIR = 'navmap'
CONFIG_FILE = 'wingman.cfg'
ROSTER_FILE = 'roster.json'
LOG_FILE = 'wingman.log'

# initialise QApplication
app = QtWidgets.QApplication([__app__.lower()])
app.setApplicationDisplayName(__app__)
app.setApplicationName(__app__.lower())
app.setWindowIcon(QtGui.QIcon(':/general/main'))

# switch current working directory to a suitable location to store app data
dataLocation = QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.AppLocalDataLocation)
os.makedirs(dataLocation, exist_ok=True)
os.chdir(dataLocation)

# configure logging
# noinspection PyArgumentList
logging.basicConfig(level=logging.INFO,
                    format='[%(asctime)s] [%(levelname)s] %(message)s',
                    handlers=[  # log to file and stdout
                            logging.FileHandler(LOG_FILE),
                            logging.StreamHandler()
                        ]
                    )
sys.excepthook = lambda *args: sys.__excepthook__(*args) if args[0] is KeyboardInterrupt \
                                                         else logging.error('Uncaught exception!', exc_info=args)

logging.info('Application start')
logging.info(f'Working directory: {os.getcwd()}')

# initialise namespaces
icons = namespaces.Icons()
config = namespaces.Configuration(CONFIG_FILE)

# platform specific initialisation
if IS_WIN:
    font = app.font()
    font.setFamily('Segoe UI')
    font.setPointSizeF(font.pointSize() * 1.2)
    app.setFont(font)
else:
    if QtWidgets.QStyleFactory.keys() == ['Windows', 'Fusion']:
        logging.warning("Native Qt style not available. Defaulting to Fusion - the application may not render "
                        "correctly. To fix this, install the PyQt5 packages from your distro's repos, not pip.")



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtWidgets


from ... import icons, __app__, __version__, __description__


class About(QtWidgets.QDialog):
    """The "About" message box."""
    title = 'About'
    vMargin = 30
    hMargin = 80

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.title)
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
        self.setWindowModality(QtCore.Qt.ApplicationModal)

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.mainLayout)
        self.mainLayout.setAlignment(QtCore.Qt.AlignHCenter)
        self.mainLayout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.mainLayout.setContentsMargins(QtCore.QMargins(self.hMargin, self.vMargin, self.hMargin, self.vMargin))

        self.iconLabel = QtWidgets.QLabel()
        self.iconLabel.setPixmap(icons.main.pixmap(128, 128))
        self.iconLabel.setAlignment(QtCore.Qt.AlignHCenter)
        self.mainLayout.addWidget(self.iconLabel, QtCore.Qt.AlignHCenter)

        self.titleLabel = QtWidgets.QLabel(f'<html><h1>{__app__} {__version__}</h1></html>')
        self.mainLayout.addWidget(self.titleLabel, QtCore.Qt.AlignHCenter)
        self.mainLayout.setAlignment(self.titleLabel, QtCore.Qt.AlignHCenter)

        self.blurbLabel = QtWidgets.QLabel(f'<html><h3>{__description__}</h3></html>')
        self.mainLayout.addWidget(self.blurbLabel, QtCore.Qt.AlignHCenter)
        self.mainLayout.setAlignment(self.blurbLabel, QtCore.Qt.AlignHCenter)

        self.authorLabel = QtWidgets.QLabel('<html><h4>made with ❤ by <a href="'
                                            'https://discoverygc.com/forums/member.php?action=profile&uid=32907">'
                                            'Biggles</a> (<a href="https://github.com/biqqles">biqqles</a>)'
                                            '</h4></html>')
        self.authorLabel.setOpenExternalLinks(True)
        self.mainLayout.addWidget(self.authorLabel, QtCore.Qt.AlignHCenter)
        self.mainLayout.setAlignment(self.authorLabel, QtCore.Qt.AlignHCenter)

        self.mainLayout.addStretch(40)

        self.licenceLabel = QtWidgets.QLabel('<html>This application is free software, released under the '
                                             '<a href="https://gnu.org/licenses/gpl.html">GNU General Public License '
                                             'v3.0</a> as is and with absolutely no warranty whatsoever.</html>')
        self.licenceLabel.setWordWrap(True)
        self.licenceLabel.setOpenExternalLinks(True)
        self.mainLayout.setAlignment(self.licenceLabel, QtCore.Qt.AlignCenter)
        self.mainLayout.addWidget(self.licenceLabel, QtCore.Qt.AlignHCenter)

        self.thanksLabel = QtWidgets.QLabel(
            '<html><h4>With thanks to:</h4><ul>'
            '<li><a href="'
            'https://discoverygc.com/forums/member.php?action=profile&uid=4734">Error</a> for the use of their '
            '<a href="https://github.com/AudunVN/Navmap">Online Navmap</a></li>'
            '<li>Icons made by:<ul>'
            '<li><a href="https://www.flaticon.com/authors/freepik">Freepik</a> from '
            '<a href="https://www.flaticon.com">www.flaticon.com</a></li>'
            '<li><a href="https://www.flaticon.com/authors/roundicons">Roundicons</a> from '
            '<a href="https://www.flaticon.com">www.flaticon.com</a></li>'
            '<li><a href="https://www.flaticon.com/authors/neungstockr">neungstockr</a> from '
            '<a href="https://www.flaticon.com">www.flaticon.com</a>.</li></ul></li>'
            '<li>See also the acknowledgements for <a href="https://github.com/biqqles/flair">flair</a> and '
            '<a href="https://github.com/biqqles/flint">flint</a>.</li>'
            '</ul></html>')
        self.thanksLabel.setOpenExternalLinks(True)
        self.mainLayout.addWidget(self.thanksLabel, QtCore.Qt.AlignHCenter)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List, Optional, Any
import os

from PyQt5 import QtCore, QtGui, QtWidgets

from .. import app, IS_WIN


class SimpleTable(QtWidgets.QTableView):
    rowSelected = QtCore.pyqtSignal('PyQt_PyObject')  # emits a list of the cells in the selected row
    rowDeselected = QtCore.pyqtSignal('PyQt_PyObject')  # emits a list of the cells in the deselected row

    def __init__(self, header: List[str]):
        super().__init__()
        # configure model
        self.itemModel = QtGui.QStandardItemModel()
        self.itemModel.setHorizontalHeaderLabels(header)
        self.setModel(self.itemModel)

        # configure view
        self.setSortingEnabled(True)
        self.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.verticalHeader().hide()
        self.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(26)  # set default row height
        self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setHighlightSections(False)  # fix for ugly/nonstandard bold header on Windows
        self.setSelectionMode(QtWidgets.QTableView.SingleSelection)

        # connections
        self.selectionModel().selectionChanged.connect(self.onSelectedRowChanged)
        self.customContextMenuRequested.connect(self.onCustomContextMenuRequested)

        if IS_WIN:
            # attempt to match the more modern styling of QTreeView and QListWidget. Hover still doesn't work, so
            # approach (e.g. QStyledItemDelegate) will be needed in future.
            self.setStyleSheet('QTableView::item:selected {'
                               '    background-color: rgb(205, 232, 255);'
                               '    color: black;'
                               '};'
                               'QTableView::item:hover {'
                               '    background-color: rgb(0, 120, 215, 20%);'
                               '};')

    def clear(self):
        """Clear the table, without clearing the headings."""
        self.itemModel.removeRows(0, self.itemModel.rowCount())

    def populate(self, rows: List[List[QtGui.QStandardItem]]):
        """Populate the table and its model with rows of items. In addition to adding this data, this method also
        readies the table for immediate use: sorting, resizing and selecting the first row."""
        self.clear()
        for row in rows:
            self.itemModel.appendRow(row)
        if rows:
            self.resizeColumnsToContents()
        self.sortByColumn(0, QtCore.Qt.AscendingOrder)
        self.selectRow(0)
        self.horizontalHeader().reset()  # fix for stretchLastSection not being obeyed sometimes

    def onSelectedRowChanged(self, selected, deselected):
        """Handle the user selecting a new row."""
        if not selected.indexes():
            return
        selectedItems = (item.data(QtCore.Qt.UserRole) for item in selected.indexes())
        self.rowSelected.emit(selectedItems)
        deselectedItems = (item.data(QtCore.Qt.UserRole) for item in deselected.indexes())
        self.rowDeselected.emit(deselectedItems)

    def onCustomContextMenuRequested(self, point):
        """Create a context menu."""
        cell = self.indexAt(point)
        menu = QtWidgets.QMenu()
        copy = menu.addMenu('Copy')
        copy.setIcon(QtGui.QIcon.fromTheme('edit-copy'))
        copy.addAction('cell').triggered.connect(lambda: self.copyToClipboard(value=cell.data()))
        copy.addAction('row').triggered.connect(lambda: self.copyToClipboard(row=cell.row()))
        copy.addAction('table').triggered.connect(self.copyToClipboard)

        menu.exec(QtGui.QCursor.pos())

    def copyToClipboard(self, row: Optional[int] = None, value: Optional[Any] = None):
        """Copy data to the clipboard as a string. If value is specified, this one value should be added. If row is
        specified, it is the index to the singular row in the model to be exported as a TSV dump. If neither are
        specified, the entire model should be dumped as TSV."""
        clipboard = app.clipboard()
        if value is not None:
            clipboard.setText(str(value))
        else:
            clipboard.setText(self.modelToTSV(row))

    def modelToTSV(self, row: Optional[int] = None):
        """Represent the item model as a TSV dump of its data. If row is specified, it is the index to the singular row
        in the model to be exported. Otherwise, all rows will be exported."""
        columns = range(self.itemModel.columnCount())
        rows = range(self.itemModel.rowCount())
        result = []
        if row is not None:
            for column in columns:
                result.append(str(self.itemModel.index(row, column).data()))
            return '\t'.join(result)
        else:
            for row in rows:
                tmp = []
                for column in columns:
                    tmp.append(str(self.itemModel.index(row, column).data()))
                result.append('\t'.join(tmp))
            return os.linesep.join(result)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtWidgets, QtCore, QtGui


class InfocardView(QtWidgets.QTextEdit):
    """A widget configured to display HTML-formatted infocards."""

    def __init__(self, parent=None):
        super().__init__()
        if parent:
            self.setParent(parent)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setReadOnly(True)

    def contextMenuEvent(self, event):
        """Show a context menu containing an action which copies the selected text to the clipboard, or the entire card if there
        isn't any."""
        selected = self.textCursor().selectedText()
        text = selected if selected else self.toPlainText()
        menu = QtWidgets.QMenu()
        copy = menu.addAction('Copy')
        copy.triggered.connect(lambda: app.clipboard().setText(text))
        copy.setIcon(QtGui.QIcon.fromTheme('edit-copy'))
        copy.setShortcut(QtGui.QKeySequence('Ctrl+C'))
        menu.exec(QtGui.QCursor.pos())


"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file contains the application's entry point - main().
"""
import logging
import sys
import os

from PyQt5 import QtWidgets
import flint as fl

from wingman import app, config, IS_WIN  # non-relative imports for the benefit of PyInstaller
from wingman.windows.main.layout import MainWindow
from wingman.windows.boxes import configuration, loading, error


def main():
    """Main application entry point."""
    if not fl.paths.is_probably_freelancer(config.paths['freelancer_dir']):
        configuration.ConfigurePaths(mandatory=True).exec()

    fl.paths.set_install_path(config.paths['freelancer_dir'])
    if IS_WIN:
        import flair
        flair.set_install_path(config.paths['freelancer_dir'])

    # display a progress bar while the game files are loaded, and an error box if loading them fails
    try:
        loading.LoadingFiles().exec()
    except (TypeError, ValueError) as e:
        logging.error(f'Failed to read game data at {config.paths["freelancer_dir"]}', exc_info=e)
        # allow the user to configure paths again
        if error.ReadError().exec() == QtWidgets.QMessageBox.AcceptRole:
            main()
            return
        sys.exit(1)

    MainWindow()

    os._exit(app.exec())  # bad! but sys.exit isn't working after QApp start for some reason


if __name__ == '__main__':
    main()



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtWidgets


class ReadError(QtWidgets.QMessageBox):
    """A message box notifying the user of a failure to read the game data."""
    def __init__(self):
        QtWidgets.QMessageBox.__init__(self)
        self.setWindowTitle('Failed to read game data')
        self.setIcon(QtWidgets.QMessageBox.Critical)
        self.setText('Please check that you have configured Wingman to use a valid Discovery Freelancer installation.')
        self.addButton(QtWidgets.QPushButton('Configure paths'), QtWidgets.QMessageBox.AcceptRole)
        self.addButton(QtWidgets.QPushButton('Exit'), QtWidgets.QMessageBox.RejectRole)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Callable, Any
import os

from PyQt5 import QtWidgets


class PathEdit(QtWidgets.QWidget):
    """A widget that allows the user to select a path, consisting of a line edit and a button to open a file picker.
    The line edit changes colour to indicate whether the path is valid or invalid."""
    def __init__(self, initialPath: str, validityCondition: Callable[[Any], bool] = lambda path: True):
        super().__init__()
        self.validPathCase = validityCondition

        self.mainLayout = QtWidgets.QHBoxLayout()
        self.setLayout(self.mainLayout)

        self.lineEdit = QtWidgets.QLineEdit()
        self.lineEdit.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        self.lineEdit.setMinimumWidth(300)
        self.lineEdit.textChanged.connect(self.onTextChanged)
        self.lineEdit.setText(initialPath)
        self.mainLayout.addWidget(self.lineEdit)

        self.browseButton = QtWidgets.QPushButton('Browse')
        self.browseButton.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        self.browseButton.clicked.connect(self.onBrowseButtonClicked)
        self.mainLayout.addWidget(self.browseButton)

        self.updateDisplayValidity()

    def onTextChanged(self, text: str):
        """Handle the text box's text being changed."""
        self.updateDisplayValidity()

    def onBrowseButtonClicked(self):
        """Open a browse dialogue which sets the path."""
        path = QtWidgets.QFileDialog.getExistingDirectory(caption='Choose folder', directory=self.lineEdit.text(),
                                                          options=QtWidgets.QFileDialog.ShowDirsOnly |
                                                          QtWidgets.QFileDialog.DontUseNativeDialog)
        if path:  # i.e. if not cancelled
            self.lineEdit.setText(path)

    def updateDisplayValidity(self):
        """Update the visuals of the PathEdit to indicate its path's current validity."""
        self.colourCode(self.isValidPath())

    def isValidPath(self) -> bool:
        """Returns whether the currently entered path is valid."""
        return self.validatePath(self.lineEdit.text())

    def validatePath(self, path) -> bool:
        """Verify a path looks correct. Some checks are always done and additional checks can be defined by defining
        `condition`.
        Todo: maybe check if directory?"""
        return os.path.exists(path) and self.validPathCase(path)

    def colourCode(self, valid):
        """Colour code the line edit to indicate whether the path is valid or invalid."""
        self.setStyleSheet('QLineEdit {{ background-color: {} }}'.format('#22bb45' if valid else '#dc143c'))



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
from setuptools import setup, find_namespace_packages


setup(
    name='wingman',  # distribution name
    version='0.1.2',

    author='biqqles',
    author_email='biqqles@protonmail.com',
    url='https://github.com/biqqles/wingman',

    description='A companion for Discovery Freelancer',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',

    packages=find_namespace_packages('src'),
    package_dir={'': 'src'},

    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Topic :: Games/Entertainment',
        'Intended Audience :: End Users/Desktop',
        'Development Status :: 3 - Alpha'
    ],

    python_requires='>=3.7',
    install_requires=[
        'dataclassy',
        'fl-flint>=0.3',
        'fl-flair; platform_system=="Windows"',
        'ago',
        'Pillow',
        'PyQt5',
        'PyQtWebEngine',
    ],

    entry_points={
          'gui_scripts': [
              'wingman = wingman.main:main'
          ]
    },
    data_files=[
        ('share/applications', ['packaging/linux/eu.biqqles.wingman.desktop']),
        ('share/icons/hicolor/256x256/apps', ['icons/general/wingman.png']),
    ],
)



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Dict, Tuple, List
from collections import defaultdict
import json
import logging
import os.path
import xml.etree.ElementTree as xml

from PyQt5 import QtCore, QtGui, QtWidgets

from .... import config, IS_WIN, ROSTER_FILE
from .layout import RosterTab
from ....models import items
from ....models.filters import TextFilter

if IS_WIN:
    import flair


class Roster:
    """The "Roster" tab provides a persistent list of accounts and their characters, along with their attributes.
    This list updates upon receiving events from flair.

    Todo: this file's code is probably the worst in the project. At the very least the model code should be split
    off."""
    def __init__(self, widget: RosterTab):
        self.widget = widget
        self.tree = self.widget.tree
        self.populateTree()
        self.serialise()

        # configure model and view
        self.sortModel = TextFilter(self.tree.model())
        self.model = self.sortModel.sourceModel()
        self.model.dataChanged.connect(self.onDataChanged)
        self.tree.setModel(self.sortModel)
        self.widget.searchLineEdit.textEdited.connect(self.onFilterTextEdited)
        self.tree.rowSelected.connect(self.onSelectedRowChanged)

        # make UI connections
        self.widget.editCharButton.clicked.connect(self.makeCharacterEditable)
        self.widget.moveCharButton.clicked.connect(self.makeCharacterMovable)
        self.widget.addCharButton.clicked.connect(lambda: self.createCharacter(self.tree.getSelectedRow()[0]))
        self.widget.removeCharButton.clicked.connect(self.removeCharacter)

        self.widget.reloadButton.clicked.connect(self.populateTree)
        self.widget.importExportButton.clicked.connect(self.export)

        if IS_WIN:
            flair.events.character_changed.connect(self.onCharacterChanged)
            flair.events.credits_changed.connect(self.onCharacterChanged)
            flair.events.system_changed.connect(self.onCharacterChanged)
            flair.events.docked.connect(self.onCharacterChanged)

    def onSelectedRowChanged(self, selection: List[QtGui.QStandardItem]):
        """Enable or disable buttons based on the type of selection."""
        isAccountRow = isinstance(selection[0], items.AccountItem)

        self.widget.addCharButton.setEnabled(isAccountRow)
        self.widget.removeCharButton.setEnabled(not isAccountRow)
        self.widget.moveCharButton.setEnabled(not isAccountRow)
        self.widget.editCharButton.setEnabled(not isAccountRow)
        self.tree.setDragEnabled(not isAccountRow)

        for item in self.tree.allItems():
            item.setEditable(False)
            item.setDragEnabled(False)

    def onFilterTextEdited(self, text):
        """Handle the filter text being changed."""
        if not text:
            self.tree.collapseAll()
        else:
            self.tree.expandAll()
        self.sortModel.update(text)

    def populateTree(self):
        """Populate the tree with deserialised data."""
        model = self.tree.model()
        self.deserialise(model)

        self.tree.reset()
        self.tree.sortByColumn(0, QtCore.Qt.AscendingOrder)
        self.tree.expandAll()

    def makeCharacterEditable(self, row=None):
        """Make the character at row editable. If row is not provided, it defaults to the selected row."""
        row = row or self.tree.getSelectedRow()
        for item in row:
            item.setEditable(True)

    def makeCharacterMovable(self, row=None):
        """Make the character at row movable between accounts. If row is not provided, it defaults to the selected
        row."""
        row = row or self.tree.getSelectedRow()
        for item in row:
            item.setDragEnabled(True)

    def removeCharacter(self):
        """Remove a character's row from the tree."""
        row = self.tree.getSelectedRow()
        row[0].parent().removeRow(row[0].row())
        self.serialise()

    def updateCharacter(self, name, account, balance=None, system=None, base=None):
        """Update a character's record."""
        model = self.model
        logged = QtCore.QDateTime.currentDateTime()

        if name is None or name == 'Trent':
            return

        # find account item
        for item in self.tree.allItems():
            if isinstance(item, items.AccountItem) and item.data(QtCore.Qt.UserRole).hash == account:
                account = item
                break
        else:
            raise ValueError(f'Account {account!r} for character {name!r} not found')

        try:
            character = self.tree.model().sourceModel().findItems(name, QtCore.Qt.MatchRecursive)[0]
        except IndexError:
            newCharacter = self.addCharacter(account, flair.state.name, flair.state.credits or 0,
                                             flair.state.system or '', flair.state.base or '', logged, '')
            logging.info(f'Added character {name!r}')
            character = newCharacter[0]
            self.tree.reset()

        else:
            index = character.index()
            if base is not None:
                model.itemFromIndex(index.siblingAtColumn(1)).putData(base)
            elif system is not None:
                model.itemFromIndex(index.siblingAtColumn(2)).putData(system)
            elif balance is not None:
                model.itemFromIndex(index.siblingAtColumn(3)).putData(balance)
            elif logged is not None:
                model.itemFromIndex(index.siblingAtColumn(4)).putData(logged)

        self.tree.scrollTo(character.index())
        self.tree.expand(character.parent().index())

    def createCharacter(self, parentAccount: items.AccountItem):
        """Create a new character and add it to the tree. Unlike addCharacter, this is intended to be triggered by the
        user."""
        newRow = self.addCharacter(parentAccount, name='', credits=0, system='', base='',
                                   logged=QtCore.QDateTime.currentDateTime().toSecsSinceEpoch(), description='')
        self.makeCharacterEditable(newRow)
        self.tree.expandAll()

    def onCharacterChanged(self, **kwargs):
        assert IS_WIN
        kwargs.update(name=flair.state.name)
        self.updateCharacter(account=flair.state.account, **kwargs)
        self.serialise()

    def onDataChanged(self, topLeft: QtCore.QModelIndex, bottomRight: QtCore.QModelIndex):
        """Handle a data edit"""
        item = self.model.itemFromIndex(topLeft)
        parent = item.parent()

        if parent is not None and parent is not self.model.invisibleRootItem():
            self.updateAccountSummary(self.model, parent)

        self.serialise()

        for column in range(self.model.columnCount()):
            self.tree.resizeColumnToContents(column)

    def export(self):
        """Allow the model to be exported to a file of the user's choice."""
        toFile = QtWidgets.QFileDialog.getSaveFileName(caption='Export characters',
                                                       directory=os.path.expanduser('~/roster.json'),
                                                       filter='JSON files (*.json)')[0]
        if toFile:
            self.serialise(toFile)

    def serialise(self, toFile=ROSTER_FILE):
        """Serialise the model to a JSON file."""
        result = {}

        for item in self.tree.allItems():  # type: GenericItem
            if item is None:
                continue

            serialised = item.serialise()
            if isinstance(item, items.AccountItem):
                account = result[serialised] = dict(characters=defaultdict(dict))
            elif isinstance(item.parent(), items.AccountItem):
                column = self.tree.itemModel.horizontalHeaderItem(item.column()).text()

                if column == 'Name':
                    character = account['characters'][serialised]
                elif column == 'Base':
                    character.update(base=serialised)
                elif column == 'System':
                    character.update(system=serialised)
                elif column == 'Credits':
                    # after a move event the model turns items back into the prototype (GenericItem). Turn them back
                    character.update(credits=items.CreditsItem(item.getData()).serialise())
                elif column == 'Last active':
                    character.update(logged=items.DateItem(item.getData()).serialise())
                elif column == 'Description':
                    character.update(description=serialised)
                else:
                    raise NotImplementedError

        with open(toFile, 'w') as f:
            json.dump(result, f, indent=2)

    @staticmethod
    def addAccount(model, code, name, description):
        """Add an account to the model, returning the new row."""
        rootItem = model.invisibleRootItem()

        accountItem = items.AccountItem(items.AccountItem.Account(code, name, description))
        newRow = [accountItem, items.BlankItem(), items.BlankItem(), items.CreditsItem(0), items.DateItem(None),
                  items.GenericItem(description)]
        rootItem.appendRow(newRow)
        return newRow

    @staticmethod
    def updateAccountSummary(model, accountItem: items.AccountItem):
        """Update an account row to summarise the attributes of the characters within."""
        balances = []
        dates = []

        for r in range(accountItem.rowCount()):
            creditsItem, dateItem = accountItem.child(r, 3), accountItem.child(r, 4)
            if None not in (creditsItem, dateItem):
                balances.append(creditsItem.getData())
                dates.append(dateItem.getData())
            # for some reason, when the row is moved these become null
            # todo: when this happens a row can be duplicated in the model if it's not edited before exit
            # todo: the old parent account also does not update

        model.itemFromIndex(accountItem.index().siblingAtColumn(3)).putData(sum(balances))
        model.itemFromIndex(accountItem.index().siblingAtColumn(4)).putData(max(dates, default=None))

    @staticmethod
    def addCharacter(parentAccount: items.AccountItem, name, credits, system, base, logged, description):
        """Add a character to the model, returning the new row."""
        dateItem = items.DateItem.deserialise(logged) if type(logged) is int else items.DateItem(logged)
        newRow = [items.GenericItem(name), items.GenericItem(base), items.GenericItem(system),
                  items.CreditsItem(credits), dateItem, items.GenericItem(description)]
        parentAccount.appendRow(newRow)
        return newRow

    @classmethod
    def deserialise(cls, model: QtGui.QStandardItemModel):
        """Deserialise roster.json, combining it with launcher accounts to create model items and adding them to the
        model. It would make more sense to return a new model here, but Qt links header names to the model so we must
        use the one the SimpleTree already has."""
        model.removeRows(0, model.rowCount() - 1)  # empty model

        if not os.path.exists(ROSTER_FILE) or os.stat(ROSTER_FILE).st_size == 0:
            db = {}
        else:
            with open(ROSTER_FILE) as f:
                db = json.load(f)

        for code, (name, description) in cls.retrieveDSLauncherAccounts().items():
            db.setdefault(code, dict(characters=defaultdict(dict)))  # create entry in db

            accountItem, *_, activityItem, descriptionItem = cls.addAccount(model, code, name, description)
            for charName, charAttributes in db[code]['characters'].items():
                cls.addCharacter(accountItem, name=charAttributes.pop('name', charName), **charAttributes)
            cls.updateAccountSummary(model, accountItem)
        return model

    @staticmethod
    def retrieveDSLauncherAccounts() -> Dict[str, Tuple[str, str]]:
        """Parse launcheraccounts.xml to a dictionary of tuples of the form {code: (name, description)}."""
        root = xml.parse(config.paths['accounts']).getroot()
        return {a.get('code'): (a.text, a.get('description')) for a in root.findall('account')}



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file contains core utility namespaces used throughout the
application.
"""
import atexit
import configparser
import logging

from PyQt5 import QtCore, QtGui, QtWidgets


class Configuration(configparser.ConfigParser):
    """A class providing a persistent configuration store. The configuration is written to storage on program exit.

    Todo: will need some way of supporting deltas on updates."""

    def __init__(self, path):
        super().__init__(self, allow_no_value=True)
        self.path = path
        self.optionxform = str  # override default behaviour of converting all keys to lowercase
        atexit.register(self.commit)

        try:
            with open(self.path) as f:
                self.read_file(f)

            self.paths = self['paths']
            self.install = self.paths['freelancer_dir']
            self.my_games = self.paths['my_games']
            self.dsace = self.paths['dsace']
            self.launcher_accounts = self.paths['accounts']

            self.urls = self['urls']
            self.navmap = self['urls']['navmap']
        except (FileNotFoundError, KeyError):
            logging.warning('Config missing or corrupt, reverting to defaults')
            self.createDefault()
            self.__init__(self.path)

    def commit(self, path=None):
        """Commit the configuration to disk."""
        path = path or self.path
        with open(path, 'w') as fh:
            self.write(fh)

    def saveAs(self, export_path: str):
        """Save/export the configuration to another file at `export_path`."""
        self.commit(export_path)

    def reset(self):
        """Reset the configuration to the defaults."""
        self.createDefault()

    def createDefault(self):
        """Create a new configuration to the defaults."""
        # load default file from resources
        file = QtCore.QFile(':/config/default')
        file.open(QtCore.QIODevice.ReadOnly | QtCore.QFile.Text)
        default = QtCore.QTextStream(file).readAll()
        self.read_string(default)
        self.commit()


class Icons:
    """Set the icons of the interface depending on whether the background colour is light (meaning that dark icons
    should be used) or dark (meaning that light icons should be used)"""
    monochrome = ['universemap', 'jump', 'expand', 'open', 'cc']

    def __init__(self):
        prefix = ':/dark/' if self.determineLuminance() else ':/light/'

        for icon in self.monochrome:
            setattr(self, icon, QtGui.QIcon(prefix + icon))

        self.main = QtGui.QIcon(':/general/main')  # todo: make static class
        self.navmap = QtGui.QIcon(':/general/navmap')
        self.merchant = QtGui.QIcon(':/general/merchant')
        self.roster = QtGui.QIcon(':/general/roster')

        # Linux only:
        self.open = QtGui.QIcon.fromTheme('document-save')
        self.copy = QtGui.QIcon.fromTheme('edit-copy')

    @staticmethod
    def determineLuminance(threshold=200) -> bool:
        """Determine whether this platform has a light or dark theme by testing the background colour of a push
        button. `threshold` is the threshold average luminance to be considered a "bright" theme."""
        button = QtWidgets.QPushButton()
        colour = button.palette().color(button.backgroundRole())
        luminance = sum(colour.getRgb()) / 3
        return luminance > threshold



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file defines the menus for the application's main window.
"""
from typing import Callable, List, Type, Set
import os.path
import subprocess

from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import QMenu, QAction, QMenuBar, QFileDialog
import flint as fl

from ..database import layout
from ..boxes import configuration, about
from ... import config, IS_WIN

if IS_WIN:
    import flair
    from flair.augment import cli, clipboard, screenshot


class SimpleAction(QAction):
    """A simplified QAction abstraction."""
    def __init__(self, title: str, run: Callable = lambda: None, shortcut: str = '', tooltip='',
                 enabled=True, checkable=False):
        super().__init__(title)
        if shortcut:
            self.setShortcut(shortcut)
        if tooltip:
            self.setToolTip(tooltip)
        if checkable:
            self.toggled.connect(run)
        else:
            self.triggered.connect(run)
        self.setEnabled(enabled)
        self.setCheckable(checkable)


class Separator(QAction):
    """A menu separator."""
    def __init__(self):
        super().__init__()
        self.setSeparator(True)


class SimpleMenu(QMenu):
    """A simplified QMenu abstraction."""
    title: str
    submenus: List['SimpleMenu'] = []
    actions_: List[SimpleAction]
    tooltip = ''

    def __init__(self, menuBar: QMenuBar):
        super().__init__(self.title, menuBar)
        self.setToolTipsVisible(True)
        for menu in self.submenus:
            self.addMenu(menu)
        self.addActions(self.actions_)


def openUrl(url: str):
    """Open a url in the default browser."""
    # noinspection PyArgumentList, PyCallByClass
    QtGui.QDesktopServices.openUrl(QtCore.QUrl(url))


def openFile(path: str):
    """Open a file in the default application."""
    # noinspection PyArgumentList, PyCallByClass
    QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(path))


def startLauncher():
    """Start the Discovery launcher in the configured installation directory."""
    subprocess.call(fl.paths.construct_path('DSLauncher.exe'), cwd=config.install)


class Utilities(SimpleMenu):
    """'Utilities' menu."""
    title = '&Utilities'
    actions_ = [
        SimpleAction('&Database', lambda: layout.Database().exec(), 'Ctrl+D'),
    ]


class File(SimpleMenu):
    """'File' menu."""
    title = '&File'

    class Browser(SimpleMenu):
        """'Open in browser' submenu."""
        title = 'Open in browser'
        actions_ = [
            SimpleAction('Server &rules', lambda: openUrl(config.urls['rules'])),
            SimpleAction('House &laws', lambda: openUrl(config.urls['houselaws'])),
            SimpleAction('&Player status', lambda: openUrl(config.urls['playerstatus'])),
            SimpleAction('Online &navmap', lambda: openUrl(config.urls['navmap'])),
            SimpleAction('&Wiki', lambda: openUrl(config.urls['wiki'])),
            SimpleAction('&Forums', lambda: openUrl(config.urls['forums'])),
        ]

    actions_ = [
        SimpleAction('&Start launcher', startLauncher, shortcut='Ctrl+L', enabled=IS_WIN),
        SimpleAction('Open DSAce.log', lambda: openFile(config.dsace)),
        SimpleAction('Open Freelancer directory', lambda: openFile(config.install)),
        Separator(),
        SimpleAction('Reload game files', shortcut='Ctrl+R'),  # todo: run
    ]

    def __init__(self, menuBar: QMenuBar):
        self.submenus = [
            self.Browser(menuBar),
        ]
        super().__init__(menuBar)


class Freelancer(SimpleMenu):
    """'Freelancer' menu."""
    title = 'Fr&eelancer'
    augmentations: Set['flair.augment.Augmentation'] = set()

    class Augmentations(SimpleMenu):
        """'Augmentations' submenu."""
        title = 'Toggle client augmentations'
        actions_ = [
            SimpleAction('Clipboard', lambda c: Freelancer.toggleAugmentation(clipboard.Clipboard, c),
                         checkable=True,
                         tooltip='Adds clipboard functionality to the chat box. '
                                 'Use Ctrl+Shift+C to copy and Ctrl+Shift+V to paste'
                         ),
            SimpleAction('Named screenshots', lambda c: Freelancer.toggleAugmentation(screenshot.Screenshot, c),
                         checkable=True,
                         tooltip="Use Ctrl+PrintScreen to take a screenshot named using the current time and"
                                 " character's name and location.\nScreenshots are saved to"
                                 " Documents/My Games/Freelancer/Screenshots"
                         ),
            SimpleAction('Command line interface', lambda c: Freelancer.toggleAugmentation(cli.CLI, c),
                         checkable=True,
                         tooltip='Adds new commands to the chat box. Send "..help" to get started'
                         ),
        ]

    actions_ = [
        SimpleAction('Bring to foreground', shortcut='Ctrl+F', enabled=IS_WIN)
    ]

    def __init__(self, menuBar):
        self.submenus = [
            self.Augmentations(menuBar)
        ]
        super().__init__(menuBar)
        if IS_WIN:
            self.actions_[0].setEnabled(flair.state.running)
            flair.events.freelancer_started.connect(lambda: self.actions_[0].setEnabled(True))
            flair.events.freelancer_stopped.connect(lambda: self.actions_[0].setEnabled(False))
            self.actions_[0].triggered.connect(flair.hook.window.make_foreground)

            if config['flair'].getboolean('clipboard'):
                self.submenus[0].actions_[0].toggle()
            if config['flair'].getboolean('screenshot'):
                self.submenus[0].actions_[1].toggle()
            if config['flair'].getboolean('cli'):
                self.submenus[0].actions_[2].toggle()
        else:
            self.submenus[0].setEnabled(False)
            self.actions_[0].setEnabled(False)

    @classmethod
    def toggleAugmentation(cls,  augmentation: Type['flair.augment.Augmentation'], toggled: bool):
        cls.loadAugmentation(augmentation) if toggled else cls.unloadAugmentation(augmentation)
        config['flair'][augmentation.__name__.lower()] = str(toggled)

    @classmethod
    def loadAugmentation(cls, augmentation: Type['flair.augment.Augmentation']):
        instance = augmentation(flair.state)
        cls.augmentations.add(instance)
        instance.load()

    @classmethod
    def unloadAugmentation(cls, augmentation: Type['flair.augment.Augmentation']):
        instance = next(filter(lambda a: isinstance(a, augmentation), cls.augmentations))
        instance.unload()


class Preferences(SimpleMenu):
    """'Preferences' menu."""
    title = '&Preferences'

    actions_ = [
        SimpleAction('Configure paths', lambda: configuration.ConfigurePaths().exec()),
        SimpleAction('Export preferences', lambda: Preferences.exportPreferences()),
        SimpleAction('Reset to defaults', config.reset),
    ]

    @staticmethod
    def exportPreferences():
        """Show a dialogue to make a copy of the preferences file."""
        path = QFileDialog.getSaveFileName(caption='Export preferences', directory=os.path.expanduser('~/wingman.cfg'),
                                           filter='Configuration files (*.cfg)',
                                           options=QFileDialog.DontUseNativeDialog)
        if path[0]:
            config.saveAs(path[0])


class Help(SimpleMenu):
    """'Help' menu."""
    title = '&Help'

    actions_ = [
        SimpleAction('Visit project thread', lambda: openUrl(config.urls['projectthread'])),
        SimpleAction('Visit project repo', lambda: openUrl(config.urls['repo'])),
        SimpleAction('About', lambda: about.About().exec()),
    ]



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
from collections import defaultdict

import flint as fl
from PyQt5 import QtWebEngineWidgets, QtCore, QtWidgets, QtGui

from ..buttons import SquareButton
from ... import config, icons, NAVMAP_DIR

DEV_JS_PATH = './widgets/mapview/navmap.js'


class MapView(QtWebEngineWidgets.QWebEngineView):
    """
    A widget that attempts to display Space's Interactive Navmap <http://space.discoverygc.com/navmap/> in such a way
    that it is seamlessly integrated into the native application.
    """
    navmapReady = QtCore.pyqtSignal(name='navmapReady')
    loadCompleted = QtCore.pyqtSignal(str, name='loadCompleted')

    # noinspection PyArgumentList
    def __init__(self, parent=None, url=config.navmap):
        """Initialise the widget."""
        QtWebEngineWidgets.QWebEngineView.__init__(self)
        self.page().profile().setCachePath(NAVMAP_DIR)  # link cache directory
        self.setHtml('<body style="background-color: transparent"></body>')  # in case the page doesn't load
        self.setUrl(QtCore.QUrl(url))
        self.setMinimumSize(512, 512)
        # Resize while maintaining aspect ratio. After much experimenting this policy and the contents of resizeEvent
        # were the best I could come up with. It's not ideal, though - it doesn't handle ExpandedMap being maximised.
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHeightForWidth(True)
        sizePolicy.setRetainSizeWhenHidden(True)
        self.setSizePolicy(sizePolicy)

        self.hide()  # hide until navmap initialised - this prevents a white flash

        if parent:
            self.setParent(parent)

        self.executeJS = self.page().runJavaScript  # create a more convenient alias

        # load JavaScript from resources, or from a file if it exists (useful for debugging)
        if os.path.isfile(DEV_JS_PATH):
            js = open(DEV_JS_PATH).read()
        else:
            jsFile = QtCore.QFile(':/javascript/navmap.js')
            if jsFile.open(QtCore.QIODevice.ReadOnly | QtCore.QFile.Text):
                js = QtCore.QTextStream(jsFile).readAll()
            else:
                raise FileNotFoundError("MapView: couldn't load hook.")

        # inject script and call initialiseNavmap when ready
        script = QtWebEngineWidgets.QWebEngineScript()
        script.setSourceCode(js)
        script.setName('navmap.js')
        script.setWorldId(QtWebEngineWidgets.QWebEngineScript.MainWorld)
        script.setInjectionPoint(QtWebEngineWidgets.QWebEngineScript.DocumentReady)
        self.page().scripts().insert(script)

        self.waitForHookLoaded()

        # create control panel
        self.controlsFrame = QtWidgets.QFrame(self)
        self.controlsFrame.setMinimumSize(30, 135)
        controlsLayout = QtWidgets.QVBoxLayout(self.controlsFrame)
        controlsLayout.setContentsMargins(0, 0, 0, 0)
        self.connInfoB = SquareButton(tooltip='Connected systems', icon=icons.jump, dropdown=True, edge=30)
        controlsLayout.addWidget(self.connInfoB)
        self.forwardB = SquareButton(tooltip='Forward', text='⭢', edge=30)
        self.forwardB.clicked.connect(self.goForward)
        controlsLayout.addWidget(self.forwardB)
        self.backB = SquareButton(tooltip='Back', text='⭠', edge=30)
        self.backB.clicked.connect(self.goBack)
        controlsLayout.addWidget(self.backB)
        self.expandB = SquareButton(tooltip='Show expanded map', icon=icons.expand, edge=30)
        controlsLayout.addWidget(self.expandB)
        self.controlsFrame.setLayout(controlsLayout)
        self.controlsFrame.show()

        # create context menu actions I tried for ages to get these options to be persistent, but among other issues
        # Qt would not emit the toggled signal when setChecked was called with the parameter set to False! In the end
        # I just dropped it, so all options will just default to ON when the application is initialised

        self.menu = QtWidgets.QMenu()

        sendState = lambda selector: lambda state: self.executeJS(
            'wingman.setState({}, {})'.format(selector, str(state).lower()))

        self.showLabels = QtWidgets.QAction('Show labels', checkable=True, checked=True)
        self.showLabels.toggled.connect(sendState('wingman.showLabels'))
        self.showZones = QtWidgets.QAction('Show nebulae', checkable=True, checked=True)
        self.showZones.toggled.connect(sendState('wingman.showZones'))
        self.showWrecks = QtWidgets.QAction('Show wrecks', checkable=True, checked=True)
        self.showWrecks.toggled.connect(sendState('wingman.showWrecks'))

        self.menu.addActions([self.showLabels, self.showWrecks, self.showZones])
        self.menu.addAction('Save map as image').triggered.connect(self.saveAsImage)

        # configure settings
        settings = self.page().view().settings()
        settings.setAttribute(QtWebEngineWidgets.QWebEngineSettings.SpatialNavigationEnabled, True)
        settings.setAttribute(QtWebEngineWidgets.QWebEngineSettings.ShowScrollBars, False)

        # initialise history
        self.backStack = []
        self.forwardsStack = []

    def initialiseNavmap(self):
        """Inject Wingman's code into the web app and change its background colour to blend in with the rest of the
        interface. """
        # Change background colour to that of window. This is the colour that will be shown when the web app is
        # loading. I found that whatever I tried, even pulling pixels(!), I could not get Qt to return the actual,
        # displayed colour of the QTabWidget (m.tw). This seems to be a known issue (see
        # <https://stackoverflow.com/a/23020483>) with no published, easy workaround. Unfortunately, we cannot set
        # the background to Qt.transparent because of a bug which prevents other widgets from being placed on top,
        # see comments at <goo.gl/pz5w5M>. n.b. if this is ever fixed, self.setAttribute(
        # QtCore.Qt.WA_TranslucentBackground) is required as well.

        # get and set background colour of background widget
        backgroundWidget = self.parentWidget() if self.parentWidget() else QtWidgets.QWidget()
        colour = backgroundWidget.palette().color(backgroundWidget.backgroundRole())
        self.page().setBackgroundColor(colour)
        self.page().runJavaScript(f'wingman.setBackgroundColour({colour.name()!r})')

        self.history().clear()  # a fresh start is always nice!

        self.urlChanged.connect(self.onUrlChange)

        if self.parentWidget():
            self.show()
        self.navmapReady.emit()

    def onUrlChange(self):
        """Handle a URL change. Wait for the variable to set, then emit loadCompleted."""
        self.page().runJavaScript('currentSystemNickname', lambda n: self.loadCompleted.emit(n) if n != 'Sirius' else 0)

        # update state of forward/backward navigation buttons and history queue
        if self.getDisplayed() and '_' not in self.getDisplayed():
            if not self.backStack or self.getDisplayed() != self.backStack[-1]:
                self.backStack.append(self.getDisplayed())

        self.backB.setEnabled(bool(len(self.backStack) - 1))
        self.forwardB.setEnabled(bool(self.forwardsStack))

    def displayEntity(self, entity: fl.entities.Entity):
        """Change the displayed object (system or solar) to the given item."""
        self.displayMap(entity.name())

    def displayMap(self, entityName: str):
        self.executeJS(f'wingman.displayMap({entityName.title()!r})')

    def displayUniverse(self):
        """Display universe map."""
        self.executeJS("wingman.displayUniverseMap()")

    def getDisplayed(self):
        """Parse URL to resolve selected object (system/base/planet) name, or 'Sirius' if the universe map is
        displayed. """
        fragment = self.url().fragment().split('&')[0]
        return fragment.replace('%20', ' ')[2:].title().replace('%27', "'") if fragment else 'Sirius'

    def setDisplayed(self, entityName):
        """Set the URL fragment to the display name of an entity, causing it to be displayed (assuming it is in the
        navmap's search array. This is useful for guaranteeing a display update, for example if Wingman's hook may have
        not loaded yet."""
        url = self.url()
        query, noClick = list(url.fragment().split('&'))
        url.setFragment(f'q={entityName}&{noClick}')
        self.setUrl(url)
        self.onUrlChange()

    def displayConnMenu(self, forSystem: fl.entities.System):
        """Add a menu of connected systems, sorted by system name. Clicking on these systems will display that
        system. """
        menu = QtWidgets.QMenu()
        menu.setToolTipsVisible(True)

        # group jumps by system
        connections = defaultdict(list)
        for jump, system in forSystem.connections().items():
            connections[system].append(jump)

        for system, jumps in sorted(connections.items(), key=lambda s_j: s_j[0].name()):
            action = menu.addAction(f'{system.name()} ({", ".join(j.type() for j in jumps)})')
            action.triggered.connect(lambda _, system=system: self.displayEntity(system))
            action.setToolTip(", ".join(j.sector() for j in jumps))
        self.connInfoB.setMenu(menu)

    def saveAsImage(self):
        """Show a dialogue allowing the user to save the currently displayed map as a PNG image."""
        defaultPath = os.path.expanduser('~/{}.png').format(self.getDisplayed())
        path = QtWidgets.QFileDialog.getSaveFileName(self,
                                                     'Save map as image',
                                                     defaultPath,
                                                     'Portable Network Graphics image (*.png)')[0]
        # Hide controls. Consider also temporarily setting the background to colour of forums (#090909) so it blends
        # in well
        self.controlsFrame.hide()
        # draw image
        image = QtGui.QImage(self.geometry().width(), self.geometry().height(), QtGui.QImage.Format_ARGB32)
        painter = QtGui.QPainter()
        painter.begin(image)
        self.page().view().render(painter)
        painter.end()

        if path:
            image.save(path)
        self.controlsFrame.show()

    def contextMenuEvent(self, event):
        """Provides a context menu that contains options to change the display settings of the map and save it as an
        image."""
        self.menu.exec(self.mapToGlobal(QtCore.QPoint(event.pos())))

    def resizeEvent(self, event):
        """Handle a resize event so that the widget stays square."""
        self.setMaximumWidth(self.height())
        self.controlsFrame.move(self.width() - 30, 0)  # keep controls stuck to top right

    def waitForHookLoaded(self, wingman=None):
        """Block until the hook has been loaded."""
        if wingman is None:
            try:
                self.page().runJavaScript('try {wingman} catch(error) {null}', self.waitForHookLoaded)
            except RuntimeError:  # thrown if process is killed
                return
        else:
            self.initialiseNavmap()

    def goForward(self):
        self.backStack.append(self.getDisplayed())
        self.displayMap(self.forwardsStack.pop())

    def goBack(self):
        self.forwardsStack.append(self.getDisplayed())
        self.backStack.pop()
        self.displayMap(self.backStack.pop())



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List, Type

from PyQt5 import QtWidgets

from ...widgets.simpletable import SimpleTable
from ...models.items import *


class DatabasePage(QtWidgets.QSplitter):
    """A page in the database, with a main table and a secondary widget which displays further information about the
    entity currently selected in the main table."""
    mainTableHeadings: List[str]

    def __init__(self, parent, infocardView, secondaryWidget):
        super().__init__(parent=parent, orientation=QtCore.Qt.Vertical)
        self.infocardView = infocardView
        self.secondaryWidget = secondaryWidget

        self.mainTable = SimpleTable(self.mainTableHeadings)
        self.mainTable.rowSelected.connect(self.onSelectedRowChanged)
        self.addWidget(self.mainTable)
        self.setStretchFactor(0, 3)

        self.addWidget(self.secondaryWidget)
        self.setStretchFactor(1, 2)
        self.populate()

        self.instance = self

    def populate(self):
        """Populate the main table with the results of a flint query."""
        raise NotImplementedError

    def filter(self):
        """Decide whether to display an Entity. Unused currently."""
        # raise NotImplementedError

    def onSelectedRowChanged(self, selectedItems):
        """Handle the selected row in the main table being changed. Subclass implementations should call this
        method so that the infocard view is updated."""
        principal, *_ = selectedItems
        self.infocardView.setText(principal.infocard())
        return principal


class BasesPage(DatabasePage):
    """Database page for bases."""
    mainTableHeadings = ['Base', 'Owner', 'System', 'Sector', 'Region', 'Base Nickname', 'System Nickname',
                         'Name ID', 'Info ID']

    def __init__(self, parent, infocardView):
        self.marketBox = QtWidgets.QGroupBox('Market')
        self.marketLayout = QtWidgets.QHBoxLayout()
        self.marketBox.setLayout(self.marketLayout)
        self.commodityTable = SimpleTable(['Commodity', 'Price', 'Sells'])
        self.equipmentTable = SimpleTable(['Equipment', 'Type', 'Price'])
        self.shipTable = SimpleTable(['Ship', 'Class', 'Package price'])
        self.marketLayout.addWidget(self.commodityTable)
        self.marketLayout.addWidget(self.equipmentTable)
        self.marketLayout.addWidget(self.shipTable)

        super().__init__(parent, infocardView, secondaryWidget=self.marketBox)

    def populate(self):
        self.mainTable.populate([
            [
                BaseItem(base),
                GenericItem(base.solar().owner().name()),
                GenericItem(base.system_().name()),
                GenericItem(base.sector()),
                GenericItem(base.system_().region()),
                MonospaceItem(base.nickname),
                MonospaceItem(base.system_().nickname),
                GenericItem(base.ids_name),
                GenericItem(base.solar().ids_info),
            ]
            for base in fl.bases if base.has_solar()]
        )

    def onSelectedRowChanged(self, selectedItems):
        base = super().onSelectedRowChanged(selectedItems)
        self.commodityTable.populate([
            [
                GenericItem(commodity.name()),
                CreditsItem(price),
                BooleanItem(commodity in base.sells_commodities()),
            ] for commodity, price in {**base.sells_commodities(), **base.buys_commodities()}.items() if commodity
        ]
        )
        self.commodityTable.sortByColumn(2, QtCore.Qt.DescendingOrder)  # sort by "sells" column

        self.equipmentTable.populate([
                [
                    GenericItem(equipment.name()),
                    GenericItem(type(equipment).__name__),
                    CreditsItem(price),
                ] for equipment, price in base.sells_equipment().items() if equipment
            ]  # todo: rep required
        )

        self.shipTable.populate([
            [
                GenericItem(ship.name()),
                GenericItem(ship.type()),
                CreditsItem(price),
            ] for ship, price in base.sells_ships().items() if ship
        ]
        )


class CommoditiesPage(DatabasePage):
    """Database page for commodities."""
    mainTableHeadings = ['Commodity', 'Default price', 'Volume', 'Decay', 'Nickname', 'Name ID', 'Info ID']

    def __init__(self, parent, infocardView):
        marketBox = QtWidgets.QGroupBox('Economy')
        marketLayout = QtWidgets.QHBoxLayout()
        marketBox.setLayout(marketLayout)
        self.economyTable = SimpleTable(['Base', 'System', 'Price', 'Sells'])
        marketLayout.addWidget(self.economyTable)

        super().__init__(parent, infocardView, secondaryWidget=marketBox)

    def populate(self):
        self.mainTable.populate([
            [
                CommodityItem(commodity),
                CreditsItem(commodity.good().price),
                NumberItem(int(commodity.volume)),
                NumberItem(int(commodity.decay_per_second)),
                MonospaceItem(commodity.nickname),
                GenericItem(commodity.ids_name),
                GenericItem(commodity.ids_info),
            ]
            for commodity in fl.commodities]
        )

    def onSelectedRowChanged(self, selectedItems):
        commodity = super().onSelectedRowChanged(selectedItems)
        sold, bought = commodity.sold_at(), commodity.bought_at()

        self.economyTable.populate([
            [
                BaseItem(base),
                SystemItem(base.system_()),
                CreditsItem(price),
                BooleanItem(base in sold)
            ]
            for base, price in {**bought, **sold}.items() if base.has_solar()
        ])
        self.economyTable.sortByColumn(3, QtCore.Qt.DescendingOrder)  # sort by "sells" column


class EquipmentPage(DatabasePage):
    """Abstract. A page for a type of equipment."""
    mainTableHeadings = ['Name', 'Price', 'Nickname', 'Name ID', 'Info ID']
    equipmentType: Type[fl.entities.Equipment] = fl.entities.Equipment

    def __init__(self, parent, infocardView):
        availabilityBox = QtWidgets.QGroupBox('Availability')
        availabilityLayout = QtWidgets.QHBoxLayout()
        availabilityBox.setLayout(availabilityLayout)
        self.economyTable = SimpleTable(['Base', 'System', 'IFF'])
        availabilityLayout.addWidget(self.economyTable)

        super().__init__(parent, infocardView, secondaryWidget=availabilityBox)

    def populate(self):
        """This base implementation populates the main table with fields common to all equipment types."""
        self.mainTable.populate([
            [
                EntityItem(equipment),
                CreditsItem(equipment.price()),
                MonospaceItem(equipment.nickname),
                GenericItem(equipment.ids_name),
                GenericItem(equipment.ids_info)
            ]
            for equipment in fl.equipment.of_type(self.equipmentType) if equipment.good()
        ])

    def onSelectedRowChanged(self, selectedItems):
        equipment = super().onSelectedRowChanged(selectedItems)
        self.economyTable.populate([
            [
                BaseItem(base),
                SystemItem(base.system_()),
                FactionItem(base.owner()),
            ] for base in equipment.sold_at() if base.has_solar()
        ])


class GunsPage(EquipmentPage):
    """Database page displaying countermeasure droppers.
    Todo: needs stats from fl.entities.Munition."""
    mainTableHeadings = ['Name', 'Price', 'Power usage', 'Refire delay', 'Muzzle velocity', 'Nickname',
                         'Name ID', 'Info ID']
    equipmentType = fl.entities.Gun

    def populate(self):
        self.mainTable.populate([
            [
                EntityItem(gun),
                CreditsItem(gun.price()),
                NumberItem(gun.power_usage),
                NumberItem(gun.refire_delay),
                NumberItem(gun.muzzle_velocity),
                MonospaceItem(gun.nickname),
                GenericItem(gun.ids_name),
                GenericItem(gun.ids_info)
            ]
            for gun in fl.equipment.of_type(self.equipmentType) if gun.good()
        ])


class ThrustersPage(EquipmentPage):
    """Database page displaying thrusters."""
    mainTableHeadings = ['Name', 'Hit points', 'Cargo space', 'Energy/s', 'Price', 'Nickname', 'Name ID', 'Info ID']
    equipmentType = fl.entities.Thruster

    def populate(self):
        self.mainTable.populate([
            [
                EntityItem(thruster),
                NumberItem(thruster.hit_pts),
                NumberItem(thruster.volume),
                NumberItem(thruster.power_usage),
                CreditsItem(thruster.price()),
                MonospaceItem(thruster.nickname),
                GenericItem(thruster.ids_name),
                GenericItem(thruster.ids_info),
            ]
            for thruster in fl.equipment.of_type(self.equipmentType) if thruster.good()
        ])


class IDsPage(EquipmentPage):
    """Database page displaying unofficial (non-serverside) IDs."""
    mainTableHeadings = ['Name', 'Nickname', 'Name ID', 'Info ID']
    equipmentType = fl.entities.Tractor

    def populate(self):
        self.mainTable.populate([
            [
                EntityItem(tractor),
                MonospaceItem(tractor.nickname),
                GenericItem(tractor.ids_name),
                GenericItem(tractor.ids_info),
            ]
            for tractor in fl.equipment.of_type(self.equipmentType) if tractor.good()
        ])


class ArmourPage(EquipmentPage):
    """Database page displaying armour upgrades."""
    mainTableHeadings = ['Name', 'Price', 'Hit points multiplier', 'Nickname', 'Name ID', 'Info ID']
    equipmentType = fl.entities.Armor

    def populate(self):
        self.mainTable.populate([
            [
                EntityItem(armour),
                CreditsItem(armour.price()),
                NumberItem(armour.hit_pts_scale),
                MonospaceItem(armour.nickname),
                GenericItem(armour.ids_name),
                GenericItem(armour.ids_info),
            ]
            for armour in fl.equipment.of_type(self.equipmentType) if armour.good()
        ])


class CountermeasuresPage(EquipmentPage):
    """Database page displaying countermeasure droppers.
    Todo: needs stats from fl.entities.CounterMeasure."""
    equipmentType = fl.entities.CounterMeasureDropper


class MinesPage(EquipmentPage):
    """Database page displaying mine droppers.
    Todo: needs stats from fl.entities.Mine."""
    equipmentType = fl.entities.MineDropper


class CloaksPage(EquipmentPage):
    """Database page displaying cloaking devices.
    Todo: needs stats."""
    equipmentType = fl.entities.CloakingDevice


class EnginesPage(EquipmentPage):
    """Database page displaying engines."""
    equipmentType = fl.entities.Engine


class ShieldsPage(EquipmentPage):
    """Database page displaying countermeasure droppers."""
    mainTableHeadings = ['Name', 'Price', 'Technology', 'Capacity', 'Resistance', 'Cargo space',
                         'Nickname', 'Name ID', 'Info ID']
    equipmentType = fl.entities.ShieldGenerator

    def populate(self):
        self.mainTable.populate([
            [
                EntityItem(shield),
                CreditsItem(shield.price()),
                GenericItem(shield.shield_type),
                NumberItem(shield.max_capacity),
                NumberItem(shield.explosion_resistance),
                NumberItem(shield.volume),
                MonospaceItem(shield.nickname),
                GenericItem(shield.ids_name),
                GenericItem(shield.ids_info),
            ]
            # temp hack to ignore shield_module_c1_fluxcoil which has the trade lane ring infocard as its name...
            for shield in fl.equipment.of_type(self.equipmentType) if 'RDL' not in shield.name()
        ])


class ShipsPage(EquipmentPage):
    """Database page displaying ships."""
    mainTableHeadings = ['Ship', 'Class', 'Price', 'Hit points', 'Hold size', 'Bots', 'Bats', 'Nickname',
                         'Name ID', 'Info ID']

    def populate(self):
        self.mainTable.populate([[
                EntityItem(ship),
                GenericItem(ship.type()),
                CreditsItem(ship.price()),
                NumberItem(ship.hit_pts),
                NumberItem(ship.hold_size),
                NumberItem(ship.nanobot_limit),
                NumberItem(ship.shield_battery_limit),
                MonospaceItem(ship.nickname),
                GenericItem(ship.ids_name),
                GenericItem(ship.ids_info),
            ] for ship in fl.ships
        ])


class FactionsPage(DatabasePage):
    """Database page displaying factions."""
    mainTableHeadings = ['Faction', 'Short name', 'Legality', 'Nickname', 'Name ID', 'Info ID']

    def __init__(self, parent, infocardView):
        self.sheetBox = QtWidgets.QGroupBox('Rep sheet')
        sheetLayout = QtWidgets.QHBoxLayout()
        self.sheetBox.setLayout(sheetLayout)
        self.sheetTable = SimpleTable(['Faction', 'Reputation towards'])
        sheetLayout.addWidget(self.sheetTable)

        super().__init__(parent, infocardView, secondaryWidget=self.sheetBox)

    def populate(self):
        self.mainTable.populate([[
                FactionItem(faction),
                GenericItem(faction.short_name()),
                GenericItem(faction.legality()),
                MonospaceItem(faction.nickname),
                GenericItem(faction.ids_name),
                GenericItem(faction.ids_info),
        ] for faction in fl.factions])

    def onSelectedRowChanged(self, selectedItems):
        """Display the currently selected faction's rep hacks and rep sheet."""
        faction = super().onSelectedRowChanged(selectedItems)
        self.sheetTable.populate([
            [
                FactionItem(other_faction),
                NumberItem(rep_towards),
            ] for other_faction, rep_towards in faction.rep_sheet().items()
        ])
        self.sheetTable.sortByColumn(1, QtCore.Qt.DescendingOrder)  # sort by "reputation with" column



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.
"""
import sys

from PyQt5 import QtCore, QtWidgets
import flint as fl


class LoadingFiles(QtWidgets.QProgressDialog):
    """Display a loading dialogue to inform the user that flint is processing the game files. This should be shown
    after paths have been configured, but before anything else."""
    class LoadingThread(QtCore.QThread):
        jobFinished = QtCore.pyqtSignal(str)

        def run(self):
            """This is a bit of a hack. Run expensive routines in flint before loading the rest of the application."""
            fl.get_systems()
            self.jobFinished.emit('Systems')
            fl.maps.generate_universe_graph()
            self.jobFinished.emit('Universe')
            fl.get_equipment()
            self.jobFinished.emit('Equipment')
            fl.get_goods()
            self.jobFinished.emit('Goods')
            fl.routines.get_markets()
            self.jobFinished.emit('Markets')

    def __init__(self):
        super().__init__()
        self.setWindowTitle('Reading game data')
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
        self.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)

        # QProgressDialog claims to "return a size that fits the contents of the progress dialog" but this is
        # absolutely not the case. As a hack, but not one that's too bad, increase its suggested size by ~50%
        self.setMinimumWidth(int(self.width() * 1.5))

        self.setRange(0, 0)
        self.setMinimumDuration(0)
        self.show()
        self.canceled.connect(lambda: sys.exit(2))

        self.thread = self.LoadingThread()
        self.thread.jobFinished.connect(self.setLabelText)
        # disconnecting canceled is required because it is emitted when the window is closed, contrary to the docs
        self.thread.finished.connect(lambda: self.canceled.disconnect() or self.close())
        self.thread.start()



"""
Copyright © 2016-2017, 2020 biqqles.

This file is part of Wingman.

Wingman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Wingman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Wingman.  If not, see <http://www.gnu.org/licenses/>.

This file defines the layout of the Navmap tab.
"""
from PyQt5 import QtCore, QtGui, QtWidgets

from ....widgets import buttons, mapview, infocardview
from .... import icons, IS_WIN


class NavmapTab(QtWidgets.QWidget):
    title = 'Navmap'
    icon = icons.navmap

    def __init__(self, parent: QtWidgets.QTabWidget):
        super().__init__(parent)
        self.mainLayout = QtWidgets.QGridLayout(self)
        self.navmap = mapview.MapView(self)
        self.navmap.setMinimumSize(QtCore.QSize(512, 512))
        self.navmap.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding))

        self.leftLayout = QtWidgets.QVBoxLayout()
        self.leftLayout.addWidget(self.navmap)

        self.mainLayout.addLayout(self.leftLayout, 0, 0)

        self.rightLayout = QtWidgets.QVBoxLayout()
        self.selectionLayout = QtWidgets.QHBoxLayout()

        self.followRadioButton = QtWidgets.QRadioButton('Follow', self)
        self.followRadioButton.setToolTip(
            'If toggled, the navmap will automatically switch to systems as you enter them')
        self.followRadioButton.setChecked(False)
        self.followRadioButton.setEnabled(False)

        self.selectionLayout.addWidget(self.followRadioButton)

        self.gotoRadioButton = QtWidgets.QRadioButton('Show: ', self)
        self.gotoRadioButton.setChecked(True)
        self.selectionLayout.addWidget(self.gotoRadioButton)

        self.searchEdit = QtWidgets.QLineEdit(self)
        self.searchEdit.setStyleSheet('font:bold')
        self.searchEdit.setPlaceholderText('Enter a system or base')
        self.searchEdit.setMinimumHeight(30)
        self.selectionLayout.addWidget(self.searchEdit)

        self.universeButton = buttons.UniverseMapButton(self)
        self.selectionLayout.addWidget(self.universeButton)

        self.rightLayout.addLayout(self.selectionLayout)

        self.infocard = infocardview.InfocardView(self)
        self.rightLayout.addWidget(self.infocard)

        self.mainLayout.addLayout(self.rightLayout, 0, 1)

